"""Tests for WebSocketDetails model."""

import pytest
from unittest.mock import AsyncMock, MagicMock

from pylamarzocco.models import WebSocketDetails


class TestWebSocketDetails:
    """Test WebSocketDetails model."""

    def test_websocket_details_init_no_params(self) -> None:
        """Test WebSocketDetails initialization with no parameters."""
        details = WebSocketDetails()
        
        assert details._ws is None
        assert details._disconnect_callback is None
        assert not details.connected

    def test_websocket_details_init_with_ws(self) -> None:
        """Test WebSocketDetails initialization with websocket."""
        mock_ws = MagicMock()
        mock_ws.closed = False
        
        details = WebSocketDetails(ws=mock_ws)
        
        assert details._ws is mock_ws
        assert details._disconnect_callback is None
        assert details.connected

    def test_websocket_details_init_with_callback(self) -> None:
        """Test WebSocketDetails initialization with disconnect callback."""
        mock_callback = AsyncMock()
        
        details = WebSocketDetails(disconnect_callback=mock_callback)
        
        assert details._ws is None
        assert details._disconnect_callback is mock_callback
        assert not details.connected

    def test_websocket_details_init_with_both(self) -> None:
        """Test WebSocketDetails initialization with both parameters."""
        mock_ws = MagicMock()
        mock_ws.closed = False
        mock_callback = AsyncMock()
        
        details = WebSocketDetails(ws=mock_ws, disconnect_callback=mock_callback)
        
        assert details._ws is mock_ws
        assert details._disconnect_callback is mock_callback
        assert details.connected

    def test_connected_property_ws_none(self) -> None:
        """Test connected property when websocket is None."""
        details = WebSocketDetails()
        
        assert not details.connected

    def test_connected_property_ws_not_closed(self) -> None:
        """Test connected property when websocket is not closed."""
        mock_ws = MagicMock()
        mock_ws.closed = False
        
        details = WebSocketDetails(ws=mock_ws)
        
        assert details.connected

    def test_connected_property_ws_closed(self) -> None:
        """Test connected property when websocket is closed."""
        mock_ws = MagicMock()
        mock_ws.closed = True
        
        details = WebSocketDetails(ws=mock_ws)
        
        assert not details.connected

    async def test_disconnect_with_callback(self) -> None:
        """Test disconnect method with callback."""
        mock_callback = AsyncMock()
        details = WebSocketDetails(disconnect_callback=mock_callback)
        
        await details.disconnect()
        
        mock_callback.assert_called_once()

    async def test_disconnect_without_callback(self) -> None:
        """Test disconnect method without callback."""
        details = WebSocketDetails()
        
        # Should not raise an exception
        await details.disconnect()

    async def test_disconnect_callback_called_once_only(self) -> None:
        """Test that disconnect callback is only called once per call."""
        mock_callback = AsyncMock()
        details = WebSocketDetails(disconnect_callback=mock_callback)
        
        await details.disconnect()
        await details.disconnect()
        
        # Should be called twice (once for each disconnect call)
        assert mock_callback.call_count == 2

    def test_websocket_state_changes(self) -> None:
        """Test websocket connection state changes."""
        mock_ws = MagicMock()
        details = WebSocketDetails(ws=mock_ws)
        
        # Initially not closed
        mock_ws.closed = False
        assert details.connected
        
        # Now closed
        mock_ws.closed = True
        assert not details.connected
        
        # Not closed again
        mock_ws.closed = False
        assert details.connected


"""Test the bluetooth client."""

import asyncio
from collections.abc import Generator
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

from pylamarzocco import LaMarzoccoBluetoothClient
from pylamarzocco.clients._bluetooth import BleakClientWithServiceCache
from pylamarzocco.const import BoilerType, MachineMode, ModelName, SmartStandByType
from pylamarzocco.exceptions import BluetoothConnectionFailed
from pylamarzocco.models import (
    BluetoothBoilerDetails,
    BluetoothMachineCapabilities,
    BluetoothSmartStandbyDetails,
)


@pytest.fixture(name="ble_device")
def ble_device_fixture() -> BLEDevice:
    """Fixture providing a fake BLE device instance."""
    return BLEDevice(
        address="test-address",
        name="Test Device",
        details=None,
    )


@pytest.fixture(name="mock_bleak_client", autouse=True)
def bleak_client() -> Generator[MagicMock, None, None]:
    """Fixture to create a mock BleakClient."""
    with patch(
        "pylamarzocco.clients._bluetooth.establish_connection",
        new_callable=AsyncMock,
    ) as mock_establish_connection:
        mock_client = MagicMock()
        mock_client.write_gatt_char = AsyncMock()
        mock_client.read_gatt_char = AsyncMock(
            return_value=b'{"id":"test-id","message":"Success","status":"success"}'
        )
        mock_client.disconnect = AsyncMock()
        mock_client.services = MagicMock()
        mock_establish_connection.return_value = mock_client
        mock_client.is_connected = True
        mock_client.services.get_characteristic.return_value = "mock_characteristic"
        mock_client.establish_mock = mock_establish_connection
        yield mock_client


SETTINGS_CHAR = "0b0b7847-e12b-09a8-b04b-8e0922a9abab"
AUTH_CHAR = "0d0b7847-e12b-09a8-b04b-8e0922a9abab"
READ_CHAR = "0a0b7847-e12b-09a8-b04b-8e0922a9abab"


async def test_ble_set_power(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test setting power on the machine."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    await client.set_power(True)

    mock_bleak_client.services.get_characteristic.assert_called_with(SETTINGS_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b'{"name":"MachineChangeMode","parameter":{"mode":"BrewingMode"}}\x00',
        response=True,
    )
    await client.disconnect()


async def test_ble_set_temperature(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test setting temperature on the machine."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    await client.set_temp(BoilerType.STEAM, 90)
    
    mock_bleak_client.services.get_characteristic.assert_called_with(SETTINGS_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b'{"name":"SettingBoilerTarget","parameter":{"identifier":"SteamBoiler","value":90}}\x00',
        response=True,
    )
    await client.disconnect()


async def test_ble_set_smart_standby(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test setting smart standby on the machine."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    await client.set_smart_standby(True, SmartStandByType.POWER_ON, 42)
    
    mock_bleak_client.services.get_characteristic.assert_called_with(SETTINGS_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b'{"name":"SettingSmartStandby","parameter":{"minutes":42,"mode":"PowerOn","enabled":true}}\x00',
        response=True,
    )
    await client.disconnect()


async def test_ble_get_machine_capability(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test getting machine capability."""
    mock_bleak_client.read_gatt_char.return_value = b'[{"family":"MICRA","groupsNumber":1,"coffeeBoilersNumber":1,"hasCupWarmer":false,"steamBoilersNumber":1,"teaDosesNumber":0,"machineModes":["BrewingMode","StandBy"],"schedulingType":"smartWakeUpSleep"}]'
    
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    response = await client.get_machine_capabilities()
    
    mock_bleak_client.services.get_characteristic.assert_called_with(READ_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b"machineCapabilities\x00",
        response=True,
    )
    assert response == BluetoothMachineCapabilities(
        family=ModelName.LINEA_MICRA,
        groups_number=1,
        coffee_boilers_number=1,
        has_cup_warmer=False,
        steam_boilers_number=1,
        tea_doses_number=0,
        machine_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
        scheduling_type="smartWakeUpSleep",
    )
    await client.disconnect()


async def test_ble_get_boiler_details(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test getting boiler details."""
    mock_bleak_client.read_gatt_char.return_value = b'[{"id":"SteamBoiler","isEnabled":true,"target":131,"current":45},{"id":"CoffeeBoiler1","isEnabled":true,"target":94,"current":65}]'
    
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    response = await client.get_boilers()
    
    mock_bleak_client.services.get_characteristic.assert_called_with(READ_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b"boilers\x00",
        response=True,
    )
    assert response == [
        BluetoothBoilerDetails(
            id=BoilerType.STEAM,
            is_enabled=True,
            target=131,
            current=45,
        ),
        BluetoothBoilerDetails(
            id=BoilerType.COFFEE,
            is_enabled=True,
            target=94,
            current=65,
        ),
    ]
    await client.disconnect()


async def test_ble_get_smart_standby_details(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test getting smart standby details."""
    mock_bleak_client.read_gatt_char.return_value = (
        b'{"mode":"PowerOn","minutes":42,"enabled":"true"}'
    )
    
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    response = await client.get_smart_standby_settings()
    
    mock_bleak_client.services.get_characteristic.assert_called_with(READ_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b"smartStandBy\x00",
        response=True,
    )
    assert response == BluetoothSmartStandbyDetails(
        mode=SmartStandByType.POWER_ON, minutes=42, enabled=True
    )
    await client.disconnect()


async def test_ble_get_tank_status(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test getting tank status."""
    mock_bleak_client.read_gatt_char.return_value = b'"true"'
    
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    response = await client.get_tank_status()
    
    mock_bleak_client.services.get_characteristic.assert_called_with(READ_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b"tankStatus\x00",
        response=True,
    )
    assert response is True
    await client.disconnect()


async def test_get_machine_mode(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test getting machine mode."""
    mock_bleak_client.read_gatt_char.return_value = b'"BrewingMode"'
    
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    response = await client.get_machine_mode()
    
    mock_bleak_client.services.get_characteristic.assert_called_with(READ_CHAR)
    mock_bleak_client.write_gatt_char.assert_called_with(
        char_specifier="mock_characteristic",
        data=b"machineMode\x00",
        response=True,
    )
    assert response == MachineMode.BREWING_MODE
    await client.disconnect()


async def test_persistent_connection_auto_connect(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that connection is established automatically on first command."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # Connection should not be established yet
    assert not client.is_connected
    
    # First command should trigger connection
    await client.set_power(True)
    
    # Connection should now be established
    mock_bleak_client.establish_mock.assert_awaited_once()
    
    # Cleanup
    await client.disconnect()


async def test_persistent_connection_reuse(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that connection is reused for multiple commands."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # Execute multiple commands
    await client.set_power(True)
    await client.set_power(False)
    await client.set_steam(True)
    
    # Connection should only be established once
    mock_bleak_client.establish_mock.assert_awaited_once()
    
    # Connection should still be active
    assert client.is_connected
    
    # Cleanup
    await client.disconnect()


async def test_auto_disconnect_after_idle(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that connection is automatically disconnected after idle timeout."""
    # Override the timeout for testing
    with patch("pylamarzocco.clients._bluetooth.IDLE_TIMEOUT", 0.1):
        client = LaMarzoccoBluetoothClient(ble_device, "token")
        
        # Execute a command to establish connection
        await client.set_power(True)
        assert client.is_connected
        
        # Wait for auto-disconnect
        await asyncio.sleep(0.2)
        
        # Connection should be closed
        assert not client.is_connected
        mock_bleak_client.disconnect.assert_awaited()


async def test_timer_reset_on_new_command(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that disconnect timer is reset when a new command is issued."""
    # Override the timeout for testing
    with patch("pylamarzocco.clients._bluetooth.IDLE_TIMEOUT", 0.2):
        client = LaMarzoccoBluetoothClient(ble_device, "token")
        
        # Execute a command
        await client.set_power(True)
        assert client.is_connected
        
        # Wait a bit but not long enough to disconnect
        await asyncio.sleep(0.1)
        
        # Execute another command (should reset timer)
        await client.set_steam(True)
        
        # Wait again
        await asyncio.sleep(0.1)
        
        # Connection should still be active (timer was reset)
        assert client.is_connected
        
        # Wait for disconnect
        await asyncio.sleep(0.2)
        assert not client.is_connected
        
        # Cleanup
        await client.disconnect()


async def test_concurrent_commands_thread_safe(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that concurrent commands are handled safely with locks."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # Execute commands concurrently
    await asyncio.gather(
        client.set_power(True),
        client.set_steam(True),
        client.set_power(False),
    )
    
    # Connection should only be established once despite concurrent calls
    mock_bleak_client.establish_mock.assert_awaited_once()
    
    # All commands should have executed
    assert mock_bleak_client.write_gatt_char.await_count >= 3
    
    # Cleanup
    await client.disconnect()


async def test_exception_triggers_disconnect(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that an exception during command execution triggers disconnect."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # First command succeeds to establish connection
    await client.set_power(True)
    assert client.is_connected
    
    # Make subsequent write fail
    mock_bleak_client.write_gatt_char.side_effect = BleakError("Connection failed")
    
    # Command should fail and trigger disconnect
    with pytest.raises(BleakError):
        await client.set_power(False)
    
    # Give time for disconnect task to complete
    await asyncio.sleep(0.05)
    
    # Disconnect should have been called
    assert not client.is_connected


async def test_characteristic_resolution_failure_clears_cache(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that failing to resolve characteristic clears cache and disconnects."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # Make characteristic resolution fail
    mock_bleak_client.services.get_characteristic.return_value = None
    mock_bleak_client.clear_cache = AsyncMock()
    
    # Command should fail
    with pytest.raises(BluetoothConnectionFailed):
        await client.set_power(True)
    
    # Cache should be cleared and disconnected
    mock_bleak_client.clear_cache.assert_awaited()
    await asyncio.sleep(0.01)  # Give time for disconnect to complete
    assert not client.is_connected


async def test_is_connected_property(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test the is_connected property."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # Should not be connected initially
    assert not client.is_connected
    
    # Connect and check
    await client.set_power(True)
    assert client.is_connected
    
    # Disconnect and check
    await client.disconnect()
    assert not client.is_connected


async def test_reconnect_after_disconnect(
    mock_bleak_client: MagicMock, ble_device: BLEDevice
) -> None:
    """Test that client can reconnect after manual disconnect."""
    client = LaMarzoccoBluetoothClient(ble_device, "token")
    
    # Connect and disconnect
    await client.set_power(True)
    await client.disconnect()
    assert not client.is_connected
    
    # Should be able to reconnect
    await client.set_steam(True)
    assert client.is_connected
    
    # Cleanup
    await client.disconnect()



"""Init for util"""
from ._authentication import (
    InstallationKey,
    generate_extra_request_headers,
    generate_installation_key,
    generate_request_proof,
)
from ._generic import is_success
from ._websocket import decode_stomp_ws_message, encode_stomp_ws_message

__all__ = [
    "InstallationKey",
    "generate_installation_key",
    "generate_extra_request_headers",
    "encode_stomp_ws_message",
    "decode_stomp_ws_message",
    "is_success",
    "generate_request_proof",
]


"""Initialize the module."""

from .clients import (
    LaMarzoccoBluetoothClient,
    LaMarzoccoCloudClient,
)

from .devices import LaMarzoccoMachine

__all__ = [
    "LaMarzoccoBluetoothClient",
    "LaMarzoccoCloudClient",
    "LaMarzoccoMachine",
]



"""Tests for websocket utility functions."""

import pytest

from pylamarzocco.const import StompMessageType
from pylamarzocco.util import decode_stomp_ws_message, encode_stomp_ws_message


class TestWebSocketUtils:
    """Test websocket utility functions."""

    def test_encode_stomp_ws_message_basic(self) -> None:
        """Test basic STOMP message encoding."""
        headers = {"destination": "/topic/test", "id": "123"}
        body = "test body"
        
        result = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, body)
        
        expected = "MESSAGE\ndestination:/topic/test\nid:123\n\ntest body\x00"
        assert result == expected

    def test_encode_stomp_ws_message_no_body(self) -> None:
        """Test STOMP message encoding without body."""
        headers = {"destination": "/topic/test"}
        
        result = encode_stomp_ws_message(StompMessageType.SUBSCRIBE, headers)
        
        expected = "SUBSCRIBE\ndestination:/topic/test\n\n\x00"
        assert result == expected

    def test_encode_stomp_ws_message_empty_headers(self) -> None:
        """Test STOMP message encoding with empty headers."""
        headers: dict[str, str] = {}
        body = "test"
        
        result = encode_stomp_ws_message(StompMessageType.CONNECT, headers, body)
        
        expected = "CONNECT\n\ntest\x00"
        assert result == expected

    def test_decode_stomp_ws_message_with_body(self) -> None:
        """Test STOMP message decoding with body."""
        message = "MESSAGE\ndestination:/topic/test\nid:123\n\ntest body\x00"
        
        msg_type, headers, body = decode_stomp_ws_message(message)
        
        assert msg_type == StompMessageType.MESSAGE
        assert headers == {"destination": "/topic/test", "id": "123"}
        assert body == "test body"

    def test_decode_stomp_ws_message_no_body(self) -> None:
        """Test STOMP message decoding without body."""
        message = "SUBSCRIBE\ndestination:/topic/test\n\n\x00"
        
        msg_type, headers, body = decode_stomp_ws_message(message)
        
        assert msg_type == StompMessageType.SUBSCRIBE
        assert headers == {"destination": "/topic/test"}
        assert body == ""

    def test_decode_stomp_ws_message_no_null_terminator(self) -> None:
        """Test STOMP message decoding without null terminator."""
        message = "CONNECT\nhost:example.com\n\nconnection data"
        
        msg_type, headers, body = decode_stomp_ws_message(message)
        
        assert msg_type == StompMessageType.CONNECT
        assert headers == {"host": "example.com"}
        assert body == "connection data"

    def test_decode_stomp_ws_message_header_with_colon(self) -> None:
        """Test STOMP message decoding with header containing colon."""
        message = "CONNECT\nauth:Bearer token:with:colons\n\n\x00"
        
        msg_type, headers, body = decode_stomp_ws_message(message)
        
        assert msg_type == StompMessageType.CONNECT
        assert headers == {"auth": "Bearer token:with:colons"}
        assert body == ""

    def test_encode_decode_round_trip(self) -> None:
        """Test encoding and decoding round trip."""
        original_headers = {
            "destination": "/ws/sn/MR123456/dashboard",
            "id": "sub-123",
            "subscription": "test-sub"
        }
        original_body = '{"test": "data", "number": 42}'
        
        # Encode
        encoded = encode_stomp_ws_message(
            StompMessageType.MESSAGE, original_headers, original_body
        )
        
        # Decode
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded)
        
        # Verify
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == original_headers
        assert decoded_body == original_body

    def test_all_stomp_message_types(self) -> None:
        """Test encoding/decoding all STOMP message types."""
        headers = {"test": "header"}
        
        for msg_type in StompMessageType:
            encoded = encode_stomp_ws_message(msg_type, headers, "body")
            decoded_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded)
            
            assert decoded_type == msg_type
            assert decoded_headers == headers
            assert decoded_body == "body"

    def test_decode_stomp_ws_message_error_type(self) -> None:
        """Test decoding ERROR message type."""
        message = "ERROR\nmessage:Invalid destination\n\nError details\x00"
        
        msg_type, headers, body = decode_stomp_ws_message(message)
        
        assert msg_type == StompMessageType.ERROR
        assert headers == {"message": "Invalid destination"}
        assert body == "Error details"


"""Models for statistics."""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any

from mashumaro import field_options
from mashumaro.mixins.json import DataClassJSONMixin

from pylamarzocco.const import DoseIndex, DoseMode, WidgetType

from ._general import BaseWidgetOutput, Thing, Widget

_LOGGER = logging.getLogger(__name__)


def _filter_valid_widgets(
    items: list[dict[str, Any] | str], field_name: str
) -> list[dict[str, Any] | str]:
    """Filter items with valid WidgetType codes, logging warnings for invalid ones.
    
    Args:
        items: List of widget dicts or code strings
        field_name: Name of the field being filtered (for logging)
    
    Returns:
        List of items with valid widget codes
    """
    valid_items = []
    for item in items:
        # Extract code - either directly if string, or from 'code' key if dict
        code = item if isinstance(item, str) else item.get("code")
        if code:
            try:
                WidgetType(code)
            except ValueError:
                # Log entire JSON if it's a dict, otherwise just the code
                if isinstance(item, dict):
                    import json
                    _LOGGER.warning(
                        "Unknown widget in field '%s' will be discarded: %s",
                        field_name,
                        json.dumps(item),
                    )
                else:
                    _LOGGER.warning(
                        "Unknown widget code '%s' in field '%s' will be discarded",
                        code,
                        field_name,
                    )
            else:
                valid_items.append(item)
    return valid_items


@dataclass(kw_only=True)
class ThingStatistics(Thing):
    """Statistics model."""

    firmwares: str | None = field(
        default=None,
    )
    selected_widget_codes: list[WidgetType] = field(
        metadata=field_options(
            alias="selectedWidgetCodes",
        ),
        default_factory=list,
    )
    all_widget_codes: list[WidgetType] = field(
        metadata=field_options(
            alias="allWidgetCodes",
        ),
        default_factory=list,
    )
    selected_widgets: list[Widget] = field(
        metadata=field_options(
            alias="selectedWidgets",
        ),
        default_factory=list,
    )

    widgets: dict[WidgetType, BaseWidgetOutput] = field(default_factory=dict)

    @classmethod
    def __pre_deserialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        # Filter out widgets with unknown codes and log warnings
        valid_widgets = _filter_valid_widgets(
            d.get("selectedWidgets", []), "selectedWidgets"
        )
        
        # Set widget_type for valid widgets
        for item in valid_widgets:
            assert isinstance(item, dict)
            item["output"]["widget_type"] = item["code"]
        
        d["selectedWidgets"] = valid_widgets
        
        # Filter widget code lists using helper
        d["selectedWidgetCodes"] = _filter_valid_widgets(
            d.get("selectedWidgetCodes", []), "selectedWidgetCodes"
        )
        d["allWidgetCodes"] = _filter_valid_widgets(
            d.get("allWidgetCodes", []), "allWidgetCodes"
        )
        
        return d

    @classmethod
    def __post_deserialize__(cls, obj: ThingStatistics) -> ThingStatistics:
        # move the firmware to a dict with type as key
        obj.widgets = {widget.code: widget.output for widget in obj.selected_widgets}
        return obj


@dataclass(kw_only=True)
class CoffeeAndFlushTrend(BaseWidgetOutput):
    """Coffee and flush trend model."""

    widget_type = WidgetType.COFFEE_AND_FLUSH_TREND
    days: int
    timezone: str
    coffees: list[CoffeeHistoryEvent]


@dataclass(kw_only=True)
class CoffeeHistoryEvent(DataClassJSONMixin):
    """Coffee that has been made in the past."""

    timestamp: datetime = field(
        metadata=field_options(
            deserialize=lambda ts: datetime.fromtimestamp(ts / 1000, tz=timezone.utc),
        )
    )
    value: int


@dataclass(kw_only=True)
class LastCoffeeList(BaseWidgetOutput):
    """Last coffees statistics."""

    widget_type = WidgetType.LAST_COFFEE
    last_coffees: list[LastCoffee] = field(
        metadata=field_options(alias="lastCoffees"),
        default_factory=list,
    )


@dataclass(kw_only=True)
class LastCoffee(DataClassJSONMixin):
    """One of the last coffees."""

    time: datetime = field(
        metadata=field_options(
            deserialize=lambda ts: datetime.fromtimestamp(ts / 1000, tz=timezone.utc),
        )
    )
    extraction_seconds: float = field(
        metadata=field_options(alias="extractionSeconds"),
    )
    dose_mode: DoseMode = field(
        metadata=field_options(alias="doseMode"),
    )
    dose_index: DoseIndex = field(
        metadata=field_options(alias="doseIndex"),
    )
    dose_value_numerator: str | None = field(
        metadata=field_options(alias="doseValueNumerator"),
        default=None,
    )


@dataclass(kw_only=True)
class CoffeeAndFlushCounter(BaseWidgetOutput):
    """Coffee counter statistics."""

    widget_type = WidgetType.COFFEE_AND_FLUSH_COUNTER
    total_coffee: int = field(
        metadata=field_options(alias="totalCoffee"),
    )
    total_flush: int = field(
        metadata=field_options(alias="totalFlush"),
    )



"""Testing the cloud client."""

from __future__ import annotations

from collections.abc import Generator
from http import HTTPMethod
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from aioresponses import aioresponses
from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1, generate_private_key
from syrupy import SnapshotAssertion
from yarl import URL

from pylamarzocco.clients import LaMarzoccoCloudClient
from pylamarzocco.const import (
    CUSTOMER_APP_URL,
    CommandStatus,
    PreExtractionMode,
    SmartStandByType,
    SteamTargetLevel,
    WeekDay,
)
from pylamarzocco.models import (
    CommandResponse,
    PrebrewSettingTimes,
    SecondsInOut,
    WakeUpScheduleSettings,
)
from pylamarzocco.util import InstallationKey

from .conftest import load_fixture

MOCK_COMMAND_RESPONSE = [
    {
        "id": "mock-id",
        "status": "Pending",
        "error_code": None,
    }
]

MOCK_SECRET_DATA = InstallationKey(
    secret=bytes(32),
    private_key=generate_private_key(SECP256R1()),
    installation_id="mock-installation-id",
)


@pytest.fixture(name="mock_ws_command_response")
def websocket_command_response() -> CommandResponse:
    """Mock websocket command response."""
    return CommandResponse(
        id="mock-id",
        status=CommandStatus.SUCCESS,
    )


@pytest.fixture(name="mock_wait_for_ws_command_response")
def wait_for_ws_command_response(
    mock_ws_command_response: CommandResponse,
) -> Generator[AsyncMock]:
    """Mock the wait for."""
    with patch(
        "pylamarzocco.clients._cloud.wait_for",
        new=AsyncMock(return_value=mock_ws_command_response),
    ) as mock_wait_for:
        yield mock_wait_for


@pytest.fixture(name="mock_websocket")
def websocket_mock() -> Generator[MagicMock]:
    """Return a mocked websocket"""
    mock_ws = MagicMock()
    mock_ws.connected = True

    with patch("pylamarzocco.clients._cloud.WebSocketDetails", return_value=mock_ws):
        yield mock_ws


async def test_access_token(mock_aioresponse: aioresponses) -> None:
    """Test getting the dashboard for a thing."""

    mock_aioresponse.post(
        url=f"{CUSTOMER_APP_URL}/auth/signin",
        status=200,
        body={
            "username": "test",
            "password": "test",
        },
        payload={
            "id": "mock-id",
            "accessToken": "mock-access",
            "refreshToken": "mock-refresh",
            "tokenType": "Bearer",
            "username": "mock-username",
            "email": "mock-email",
        },
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.async_get_access_token()
    assert result == "mock-access"

    # now get one again to get from cache
    mock_aioresponse.post(
        url=f"{CUSTOMER_APP_URL}/auth/signin",
        status=200,
        payload={
            "accessToken": "new-new-token",
            "refreshToken": "mock-refresh",
        },
    )
    result = await client.async_get_access_token()
    assert result == "mock-access"

    # now get one from refresh token
    mock_aioresponse.post(
        url=f"{CUSTOMER_APP_URL}/auth/refreshtoken",
        body={
            "username": "test",
            "refreshToken": "mock-refresh",
        },
        payload={
            "accessToken": "new-token",
            "refreshToken": "new-refresh",
            "tokenType": "Bearer",
        },
    )

    with patch("pylamarzocco.clients._cloud.TOKEN_TIME_TO_REFRESH", new=432001):
        result = await client.async_get_access_token()

    assert result == "new-token"


@pytest.mark.parametrize("model", ["micra", "gs3av", "mini", "minir"])
async def test_get_thing_dashboard(
    mock_aioresponse: aioresponses,
    model: str,
    serial: str,
    snapshot: SnapshotAssertion,
) -> None:
    """Test getting the dashboard for a thing."""

    mock_aioresponse.get(
        url=f"{CUSTOMER_APP_URL}/things/{serial}/dashboard",
        status=200,
        payload=load_fixture("machine", f"dashboard_{model}.json"),
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.get_thing_dashboard(serial)
    assert result.to_dict() == snapshot


async def test_get_thing_settings(
    mock_aioresponse: aioresponses, serial: str, snapshot: SnapshotAssertion
) -> None:
    """Test getting the settings for a thing."""

    mock_aioresponse.get(
        url=f"{CUSTOMER_APP_URL}/things/{serial}/settings",
        status=200,
        payload=load_fixture("machine", "settings_micra.json"),
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.get_thing_settings(serial)
    assert result.to_dict() == snapshot


async def test_get_thing_schedule(
    mock_aioresponse: aioresponses, serial: str, snapshot: SnapshotAssertion
) -> None:
    """Test getting the schedule for a thing."""

    mock_aioresponse.get(
        url=f"{CUSTOMER_APP_URL}/things/{serial}/scheduling",
        status=200,
        payload=load_fixture("machine", "schedule.json"),
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.get_thing_schedule(serial)
    assert result.to_dict() == snapshot


async def test_list_things(
    mock_aioresponse: aioresponses, snapshot: SnapshotAssertion
) -> None:
    """Test getting the list of things."""

    mock_aioresponse.get(
        url=f"{CUSTOMER_APP_URL}/things",
        status=200,
        payload=[load_fixture("machine", "settings_micra.json")],
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.list_things()
    assert result[0].to_dict() == snapshot


async def test_get_statistics(
    mock_aioresponse: aioresponses, serial: str, snapshot: SnapshotAssertion
) -> None:
    """Test getting the list of things."""

    mock_aioresponse.get(
        url=f"{CUSTOMER_APP_URL}/things/{serial}/stats",
        status=200,
        payload=load_fixture("machine", "statistics.json"),
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.get_thing_statistics(serial)
    assert result.to_dict() == snapshot


async def test_set_power(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the power for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineChangeMode"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)

    result = await client.set_power(serial, False)

    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {"mode": "StandBy"}
    assert result is True


@pytest.mark.usefixtures("mock_websocket", "mock_wait_for_ws_command_response")
async def test_set_power_with_ws_validation(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the power for a thing, validate the command from ws."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineChangeMode"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)

    result = await client.set_power(serial, False)

    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {"mode": "StandBy"}
    assert result is True


@pytest.mark.usefixtures("mock_websocket", "mock_wait_for_ws_command_response")
async def test_failing_response_ws_validation(
    mock_aioresponse: aioresponses,
    mock_ws_command_response: CommandResponse,
    serial: str,
) -> None:
    """Tests failing response from websocket"""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineChangeMode"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    mock_ws_command_response.status = CommandStatus.ERROR

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)

    result = await client.set_power(serial, False)
    assert result is False


@pytest.mark.usefixtures("mock_websocket", "mock_wait_for_ws_command_response")
async def test_pending_command_ws_validation_timeout(
    mock_aioresponse: aioresponses,
    mock_wait_for_ws_command_response: AsyncMock,
    serial: str,
) -> None:
    """Tests failing response from websocket"""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineChangeMode"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    mock_wait_for_ws_command_response.side_effect = TimeoutError

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)

    result = await client.set_power(serial, False)
    assert result is False


async def test_disconnected_ws_returns_true(
    mock_aioresponse: aioresponses,
    mock_websocket: MagicMock,
    serial: str,
) -> None:
    """Test setting the power for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineChangeMode"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)

    mock_websocket.connected = False

    result = await client.set_power(serial, False)

    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {"mode": "StandBy"}
    assert result is True


async def test_set_steam(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the steam for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineSettingSteamBoilerEnabled"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.set_steam(serial, True)

    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "boilerIndex": 1,
        "enabled": True,
    }
    assert result is True


async def test_set_coffee_temperature(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the steam for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineSettingCoffeeBoilerTargetTemperature"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.set_coffee_target_temperature(serial, 94.584)

    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "boilerIndex": 1,
        "targetTemperature": 94.6,
    }
    assert result is True


async def test_set_steam_target_level(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the steam target level for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineSettingSteamBoilerTargetLevel"

    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.set_steam_target_level(serial, SteamTargetLevel.LEVEL_1)
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "boilerIndex": 1,
        "targetLevel": "Level1",
    }
    assert result is True


async def test_start_backflush_cleaning(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test starting backflush cleaning for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineBackFlushStartCleaning"
    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.start_backflush_cleaning(serial)
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "enabled": True,
    }
    assert result is True


async def test_change_pre_extraction_mode(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test changing the pre-extraction mode for a thing."""

    url = (
        f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachinePreBrewingChangeMode"
    )
    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.change_pre_extraction_mode(
        serial, PreExtractionMode.PREBREWING
    )
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "mode": "PreBrewing",
    }
    assert result is True


async def test_change_pre_extraction_times(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test changing the pre-extraction times for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachinePreBrewingSettingTimes"
    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.change_pre_extraction_times(
        serial,
        PrebrewSettingTimes(times=SecondsInOut(seconds_in=5.12, seconds_out=5.03)),
    )
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "times": {"In": 5.1, "Out": 5.0},
        "groupIndex": 1,
        "doseIndex": "ByGroup",
    }
    assert result is True


async def test_setting_smart_standby(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the smart standby for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineSettingSmartStandBy"
    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.set_smart_standby(
        serial, False, 20, SmartStandByType.LAST_BREW
    )
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    assert call.kwargs["json"] == {
        "enabled": False,
        "minutes": 20,
        "after": "LastBrewing",
    }
    assert result is True


async def test_set_wake_up_schedule(
    mock_aioresponse: aioresponses,
    serial: str,
) -> None:
    """Test setting the wake up schedule for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/command/CoffeeMachineSetWakeUpSchedule"
    mock_aioresponse.post(
        url=url,
        status=200,
        payload=MOCK_COMMAND_RESPONSE,
        repeat=2,
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)

    # new schedule
    result = await client.set_wakeup_schedule(
        serial,
        WakeUpScheduleSettings(
            enabled=True,
            on_time_minutes=50,
            off_time_minutes=1439,
            steam_boiler=False,
            days=[WeekDay.MONDAY, WeekDay.FRIDAY],
        ),
    )
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][0]
    expected_output = {
        "enabled": True,
        "onTimeMinutes": 50,
        "offTimeMinutes": 1439,
        "days": [
            "Monday",
            "Friday",
        ],
        "steamBoiler": False,
    }
    assert call.kwargs["json"] == expected_output
    assert result is True

    # existing schedule
    result = await client.set_wakeup_schedule(
        serial,
        WakeUpScheduleSettings(
            identifier="aBc23d",
            enabled=True,
            on_time_minutes=50,
            off_time_minutes=1439,
            steam_boiler=False,
            days=[WeekDay.MONDAY, WeekDay.FRIDAY],
        ),
    )
    call = mock_aioresponse.requests[(HTTPMethod.POST, URL(url))][1]
    assert call.kwargs["json"] == {
        "id": "aBc23d",
        **expected_output,
    }
    assert result is True


async def test_get_update_details(
    mock_aioresponse: aioresponses, serial: str, snapshot: SnapshotAssertion
) -> None:
    """Test getting the update details for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/update-fw"
    mock_aioresponse.get(
        url=url,
        status=200,
        payload={
            "status": "ToUpdate",
            "commandStatus": "InProgress",
            "progressInfo": "download",
            "progressPercentage": 10,
        },
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.get_thing_firmware(serial)
    assert result.to_dict() == snapshot


async def test_start_update(
    mock_aioresponse: aioresponses, serial: str, snapshot: SnapshotAssertion
) -> None:
    """Test getting the update details for a thing."""

    url = f"{CUSTOMER_APP_URL}/things/{serial}/update-fw"
    mock_aioresponse.post(
        url=url,
        status=200,
        payload={
            "status": "ToUpdate",
            "commandStatus": "InProgress",
            "progressInfo": "starting process",
            "progressPercentage": None,
        },
    )

    client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
    result = await client.update_firmware(serial)
    assert result.to_dict() == snapshot



"""Test Bluetooth dashboard functionality."""

from datetime import datetime, timezone
from typing import cast
from unittest.mock import AsyncMock, MagicMock

import pytest
from syrupy.assertion import SnapshotAssertion

from pylamarzocco import LaMarzoccoBluetoothClient, LaMarzoccoMachine
from pylamarzocco.const import (
    BoilerStatus,
    BoilerType,
    MachineMode,
    MachineState,
    ModelCode,
    ModelName,
    SteamTargetLevel,
    WidgetType,
)
from pylamarzocco.exceptions import BluetoothConnectionFailed
from pylamarzocco.models import (
    BluetoothBoilerDetails,
    BluetoothCommandStatus,
    BluetoothMachineCapabilities,
    CoffeeBoiler,
    MachineStatus,
    SteamBoilerLevel,
    SteamBoilerTemperature,
)


@pytest.fixture(name="mock_bluetooth_client")
def mock_lm_bluetooth_client() -> MagicMock:
    """Mock the LaMarzoccoBluetoothClient."""
    client = MagicMock(spec=LaMarzoccoBluetoothClient)
    return client


@pytest.fixture(name="mock_machine_with_dashboard")
def mock_lm_machine_with_dashboard(
    mock_bluetooth_client: MagicMock,
) -> LaMarzoccoMachine:
    """Mock the LaMarzoccoMachine with dashboard populated."""
    machine = LaMarzoccoMachine(
        serial_number="MR123456",
        bluetooth_client=mock_bluetooth_client,
    )

    # Mock the connection_date to a fixed value for snapshot tests
    machine.dashboard.connection_date = datetime(2024, 1, 1, 12, 0, 0, tzinfo=timezone.utc)

    # Set up dashboard config with widgets
    machine.dashboard.config = {
        WidgetType.CM_MACHINE_STATUS: MachineStatus(
            status=MachineState.STANDBY,
            available_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
            mode=MachineMode.STANDBY,
            next_status=None,
            brewing_start_time=None,
        ),
        WidgetType.CM_COFFEE_BOILER: CoffeeBoiler(
            status=BoilerStatus.STAND_BY,
            enabled=True,
            enabled_supported=False,
            target_temperature=93.0,
            target_temperature_min=80,
            target_temperature_max=100,
            target_temperature_step=0.1,
            ready_start_time=None,
        ),
        WidgetType.CM_STEAM_BOILER_LEVEL: SteamBoilerLevel(
            status=BoilerStatus.STAND_BY,
            enabled=True,
            enabled_supported=True,
            target_level=SteamTargetLevel.LEVEL_1,
            target_level_supported=True,
            ready_start_time=None,
        ),
        WidgetType.CM_STEAM_BOILER_TEMPERATURE: SteamBoilerTemperature(
            status=BoilerStatus.STAND_BY,
            enabled=True,
            enabled_supported=False,
            target_temperature=126.0,
            target_temperature_min=126,
            target_temperature_max=131,
            target_temperature_step=1.0,
            target_temperature_supported=True,
            ready_start_time=None,
        ),
    }

    return machine


async def test_get_dashboard_from_bluetooth(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
    snapshot: SnapshotAssertion,
) -> None:
    """Test filling dashboard from Bluetooth."""
    # Set up mock responses
    mock_bluetooth_client.get_machine_capabilities = AsyncMock(
        return_value=BluetoothMachineCapabilities(
            family=ModelName.LINEA_MICRA,
            groups_number=1,
            coffee_boilers_number=1,
            has_cup_warmer=False,
            steam_boilers_number=1,
            tea_doses_number=0,
            machine_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
            scheduling_type="weekly",
        )
    )
    mock_bluetooth_client.get_machine_mode = AsyncMock(
        return_value=MachineMode.BREWING_MODE
    )
    mock_bluetooth_client.get_boilers = AsyncMock(
        return_value=[
            BluetoothBoilerDetails(
                id=BoilerType.COFFEE,
                is_enabled=True,
                target=95,
                current=94,
            ),
            BluetoothBoilerDetails(
                id=BoilerType.STEAM,
                is_enabled=False,
                target=128,
                current=50,
            ),
        ]
    )

    await mock_machine_with_dashboard.get_dashboard_from_bluetooth()

    # Verify calls - capabilities should NOT be called automatically anymore
    mock_bluetooth_client.get_machine_capabilities.assert_not_called()
    mock_bluetooth_client.get_machine_mode.assert_called_once()
    mock_bluetooth_client.get_boilers.assert_called_once()

    # Snapshot test includes model_name, model_code, and config
    assert mock_machine_with_dashboard.dashboard.to_dict() == snapshot

    # Verify dashboard was updated
    machine_status = cast(
        MachineStatus,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_MACHINE_STATUS],
    )
    assert machine_status.mode == MachineMode.BREWING_MODE

    coffee_boiler = cast(
        CoffeeBoiler,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_COFFEE_BOILER],
    )
    assert coffee_boiler.enabled is True
    assert coffee_boiler.target_temperature == 95.0

    # MICRA only has steam level widget, not temperature
    steam_level = cast(
        SteamBoilerLevel,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_STEAM_BOILER_LEVEL],
    )
    assert steam_level.enabled is False

    # MICRA should NOT have steam temperature widget
    assert (
        WidgetType.CM_STEAM_BOILER_TEMPERATURE
        not in mock_machine_with_dashboard.dashboard.config
    )


async def test_get_dashboard_no_bluetooth(
    mock_machine_with_dashboard: LaMarzoccoMachine,
) -> None:
    """Test filling dashboard without Bluetooth client."""
    mock_machine_with_dashboard._bluetooth_client = None  # pylint:disable=W0212

    with pytest.raises(BluetoothConnectionFailed):
        await mock_machine_with_dashboard.get_dashboard_from_bluetooth()


async def test_get_model_info_from_bluetooth(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test fetching model information from Bluetooth."""
    # Set up mock responses
    mock_bluetooth_client.get_machine_capabilities = AsyncMock(
        return_value=BluetoothMachineCapabilities(
            family=ModelName.GS3,
            groups_number=3,
            coffee_boilers_number=2,
            has_cup_warmer=True,
            steam_boilers_number=1,
            tea_doses_number=0,
            machine_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
            scheduling_type="weekly",
        )
    )

    # Call the method to fetch model info
    await mock_machine_with_dashboard.get_model_info_from_bluetooth()

    # Verify capabilities were called
    mock_bluetooth_client.get_machine_capabilities.assert_called_once()

    # Verify model info was set correctly
    assert mock_machine_with_dashboard.dashboard.model_name == ModelName.GS3
    assert mock_machine_with_dashboard.dashboard.model_code == ModelCode.GS3


async def test_get_dashboard_initializes_missing_widgets(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
    snapshot: SnapshotAssertion,
) -> None:
    """Test that get_dashboard_from_bluetooth initializes widgets if they don't exist."""
    # Clear the dashboard to simulate widgets not existing
    mock_machine_with_dashboard.dashboard.config.clear()

    # Set up mock responses
    mock_bluetooth_client.get_machine_capabilities = AsyncMock(
        return_value=BluetoothMachineCapabilities(
            family=ModelName.LINEA_MINI_R,
            groups_number=1,
            coffee_boilers_number=1,
            has_cup_warmer=True,
            steam_boilers_number=1,
            tea_doses_number=0,
            machine_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
            scheduling_type="weekly",
        )
    )
    mock_bluetooth_client.get_machine_mode = AsyncMock(
        return_value=MachineMode.BREWING_MODE
    )
    mock_bluetooth_client.get_boilers = AsyncMock(
        return_value=[
            BluetoothBoilerDetails(
                id=BoilerType.COFFEE,
                is_enabled=True,
                target=92,
                current=91,
            ),
            BluetoothBoilerDetails(
                id=BoilerType.STEAM,
                is_enabled=True,
                target=130,
                current=125,
            ),
        ]
    )

    # First fetch model info explicitly
    await mock_machine_with_dashboard.get_model_info_from_bluetooth()

    # Then get dashboard
    await mock_machine_with_dashboard.get_dashboard_from_bluetooth()

    # Verify model_name and model_code were set from capabilities
    assert mock_machine_with_dashboard.dashboard.model_name == ModelName.LINEA_MINI_R
    assert mock_machine_with_dashboard.dashboard.model_code == ModelCode.LINEA_MINI_R

    # Verify widgets were created (Linea Mini R supports steam level, not temperature)
    assert WidgetType.CM_MACHINE_STATUS in mock_machine_with_dashboard.dashboard.config
    assert WidgetType.CM_COFFEE_BOILER in mock_machine_with_dashboard.dashboard.config
    assert (
        WidgetType.CM_STEAM_BOILER_LEVEL in mock_machine_with_dashboard.dashboard.config
    )
    assert (
        WidgetType.CM_STEAM_BOILER_TEMPERATURE
        not in mock_machine_with_dashboard.dashboard.config
    )

    # Verify widgets have correct values from Bluetooth
    assert mock_machine_with_dashboard.dashboard.to_dict() == snapshot


async def test_get_dashboard_without_steam_level_support(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that models without steam level support only get temperature widget."""
    # Clear the dashboard to simulate widgets not existing
    mock_machine_with_dashboard.dashboard.config.clear()

    # Set up mock responses for GS3 (does not support steam level)
    mock_bluetooth_client.get_machine_capabilities = AsyncMock(
        return_value=BluetoothMachineCapabilities(
            family=ModelName.GS3,
            groups_number=3,
            coffee_boilers_number=2,
            has_cup_warmer=True,
            steam_boilers_number=1,
            tea_doses_number=0,
            machine_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
            scheduling_type="weekly",
        )
    )
    mock_bluetooth_client.get_machine_mode = AsyncMock(
        return_value=MachineMode.BREWING_MODE
    )
    mock_bluetooth_client.get_boilers = AsyncMock(
        return_value=[
            BluetoothBoilerDetails(
                id=BoilerType.COFFEE,
                is_enabled=True,
                target=94,
                current=93,
            ),
            BluetoothBoilerDetails(
                id=BoilerType.STEAM,
                is_enabled=True,
                target=128,
                current=120,
            ),
        ]
    )

    # First fetch model info explicitly
    await mock_machine_with_dashboard.get_model_info_from_bluetooth()

    # Then get dashboard
    await mock_machine_with_dashboard.get_dashboard_from_bluetooth()

    # Verify model_name and model_code were set from capabilities
    assert mock_machine_with_dashboard.dashboard.model_name == ModelName.GS3
    assert mock_machine_with_dashboard.dashboard.model_code == ModelCode.GS3

    # Verify widgets were created - GS3 should NOT have steam level widget
    assert WidgetType.CM_MACHINE_STATUS in mock_machine_with_dashboard.dashboard.config
    assert WidgetType.CM_COFFEE_BOILER in mock_machine_with_dashboard.dashboard.config
    assert (
        WidgetType.CM_STEAM_BOILER_LEVEL
        not in mock_machine_with_dashboard.dashboard.config
    )
    assert (
        WidgetType.CM_STEAM_BOILER_TEMPERATURE
        in mock_machine_with_dashboard.dashboard.config
    )

    # Verify temperature widget has correct values
    steam_temp = cast(
        SteamBoilerTemperature,
        mock_machine_with_dashboard.dashboard.config[
            WidgetType.CM_STEAM_BOILER_TEMPERATURE
        ],
    )
    assert steam_temp.enabled is True
    assert steam_temp.target_temperature == 128.0


async def test_get_dashboard_mini_original_temperature_only(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that original Linea Mini (without R) only gets temperature widget."""
    # Clear the dashboard to simulate widgets not existing
    mock_machine_with_dashboard.dashboard.config.clear()

    # Set up mock responses for original Linea Mini
    mock_bluetooth_client.get_machine_capabilities = AsyncMock(
        return_value=BluetoothMachineCapabilities(
            family=ModelName.LINEA_MINI,
            groups_number=1,
            coffee_boilers_number=1,
            has_cup_warmer=False,
            steam_boilers_number=1,
            tea_doses_number=0,
            machine_modes=[MachineMode.BREWING_MODE, MachineMode.STANDBY],
            scheduling_type="weekly",
        )
    )
    mock_bluetooth_client.get_machine_mode = AsyncMock(
        return_value=MachineMode.BREWING_MODE
    )
    mock_bluetooth_client.get_boilers = AsyncMock(
        return_value=[
            BluetoothBoilerDetails(
                id=BoilerType.COFFEE,
                is_enabled=True,
                target=93,
                current=92,
            ),
            BluetoothBoilerDetails(
                id=BoilerType.STEAM,
                is_enabled=True,
                target=127,
                current=115,
            ),
        ]
    )

    # First fetch model info explicitly
    await mock_machine_with_dashboard.get_model_info_from_bluetooth()

    # Then get dashboard
    await mock_machine_with_dashboard.get_dashboard_from_bluetooth()

    # Verify model_name and model_code were set from capabilities
    assert mock_machine_with_dashboard.dashboard.model_name == ModelName.LINEA_MINI
    assert mock_machine_with_dashboard.dashboard.model_code == ModelCode.LINEA_MINI

    # Verify widgets - original Mini should NOT have steam level, only temperature
    assert WidgetType.CM_MACHINE_STATUS in mock_machine_with_dashboard.dashboard.config
    assert WidgetType.CM_COFFEE_BOILER in mock_machine_with_dashboard.dashboard.config
    assert (
        WidgetType.CM_STEAM_BOILER_LEVEL
        not in mock_machine_with_dashboard.dashboard.config
    )
    assert (
        WidgetType.CM_STEAM_BOILER_TEMPERATURE
        in mock_machine_with_dashboard.dashboard.config
    )

    # Verify temperature widget has correct values
    steam_temp = cast(
        SteamBoilerTemperature,
        mock_machine_with_dashboard.dashboard.config[
            WidgetType.CM_STEAM_BOILER_TEMPERATURE
        ],
    )
    assert steam_temp.enabled is True
    assert steam_temp.target_temperature == 127.0


async def test_set_power_updates_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
    snapshot: SnapshotAssertion,
) -> None:
    """Test that set_power updates dashboard on success."""
    mock_bluetooth_client.set_power = AsyncMock(
        return_value=BluetoothCommandStatus(
            id="ble", message="power on", status="success"
        )
    )

    result = await mock_machine_with_dashboard.set_power(True)

    assert result is True
    machine_status = cast(
        MachineStatus,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_MACHINE_STATUS],
    )
    assert machine_status.mode == MachineMode.BREWING_MODE
    assert machine_status.to_dict() == snapshot


async def test_set_power_off_updates_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that set_power(False) updates dashboard on success."""
    # First set to brewing mode
    machine_status_temp = cast(
        MachineStatus,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_MACHINE_STATUS],
    )
    machine_status_temp.mode = MachineMode.BREWING_MODE

    mock_bluetooth_client.set_power = AsyncMock(
        return_value=BluetoothCommandStatus(
            id="ble", message="power on", status="success"
        )
    )

    result = await mock_machine_with_dashboard.set_power(False)

    assert result is True
    machine_status = cast(
        MachineStatus,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_MACHINE_STATUS],
    )
    assert machine_status.mode == MachineMode.STANDBY


async def test_set_steam_updates_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that set_steam updates dashboard on success."""
    mock_bluetooth_client.set_steam = AsyncMock(
        return_value=BluetoothCommandStatus(
            id="ble", message="boiler enable success", status="success"
        )
    )

    result = await mock_machine_with_dashboard.set_steam(False)

    assert result is True
    steam_level = cast(
        SteamBoilerLevel,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_STEAM_BOILER_LEVEL],
    )
    assert steam_level.enabled is False

    steam_temp = cast(
        SteamBoilerTemperature,
        mock_machine_with_dashboard.dashboard.config[
            WidgetType.CM_STEAM_BOILER_TEMPERATURE
        ],
    )
    assert steam_temp.enabled is False


async def test_set_coffee_temp_updates_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
    snapshot: SnapshotAssertion,
) -> None:
    """Test that set_coffee_target_temperature updates dashboard on success."""
    mock_bluetooth_client.set_temp = AsyncMock(
        return_value=BluetoothCommandStatus(
            id="ble", message="Setting Temperature Success", status="success"
        )
    )

    result = await mock_machine_with_dashboard.set_coffee_target_temperature(96.5)

    assert result is True
    coffee_boiler = cast(
        CoffeeBoiler,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_COFFEE_BOILER],
    )
    assert coffee_boiler.target_temperature == 96.5
    assert coffee_boiler.to_dict() == snapshot


async def test_set_steam_temp_updates_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that set_steam_level updates dashboard on success."""
    mock_bluetooth_client.set_temp = AsyncMock(
        return_value=BluetoothCommandStatus(
            id="ble", message="Setting Temperature Success", status="success"
        )
    )

    # Call set_steam_level which uses set_temp for steam boiler
    result = await mock_machine_with_dashboard.set_steam_level(SteamTargetLevel.LEVEL_3)

    assert result is True
    steam_temp = cast(
        SteamBoilerTemperature,
        mock_machine_with_dashboard.dashboard.config[
            WidgetType.CM_STEAM_BOILER_TEMPERATURE
        ],
    )
    assert steam_temp.target_temperature == 131.0


async def test_failed_command_does_not_update_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that failed commands don't update dashboard."""
    machine_status_orig = cast(
        MachineStatus,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_MACHINE_STATUS],
    )
    original_mode = machine_status_orig.mode

    mock_bluetooth_client.set_power = AsyncMock(
        return_value=BluetoothCommandStatus(id="ble", message="Failed", status="error")
    )

    result = await mock_machine_with_dashboard.set_power(True)

    # Command returns False but doesn't raise exception
    assert result is False
    # Dashboard should not be updated
    machine_status = cast(
        MachineStatus,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_MACHINE_STATUS],
    )
    assert machine_status.mode == original_mode


async def test_bluetooth_exception_does_not_update_dashboard(
    mock_machine_with_dashboard: LaMarzoccoMachine,
    mock_bluetooth_client: MagicMock,
) -> None:
    """Test that Bluetooth exceptions don't update dashboard."""
    coffee_boiler_orig = cast(
        CoffeeBoiler,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_COFFEE_BOILER],
    )
    original_temp = coffee_boiler_orig.target_temperature

    mock_bluetooth_client.set_temp = AsyncMock(
        side_effect=BluetoothConnectionFailed("Connection lost")
    )

    # This will fail and return False (no cloud client)
    result = await mock_machine_with_dashboard.set_coffee_target_temperature(96.5)

    assert result is False
    # Dashboard should not be updated
    coffee_boiler = cast(
        CoffeeBoiler,
        mock_machine_with_dashboard.dashboard.config[WidgetType.CM_COFFEE_BOILER],
    )
    assert coffee_boiler.target_temperature == original_temp



"""Tests for websocket message handling in cloud client."""

import pytest
from unittest.mock import MagicMock, patch, AsyncMock
from aiohttp import WSMessage, WSMsgType

from pylamarzocco.clients import LaMarzoccoCloudClient
from pylamarzocco.const import StompMessageType
from pylamarzocco.util import InstallationKey, encode_stomp_ws_message
from pylamarzocco.models import WebSocketDetails
from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1, generate_private_key


MOCK_SECRET_DATA = InstallationKey(
    secret=bytes(32),
    private_key=generate_private_key(SECP256R1()),
    installation_id="mock-installation-id",
)


class TestWebSocketMessageHandling:
    """Test websocket message handling functionality."""

    @pytest.fixture
    async def mock_client(self):
        """Create a mocked client with session."""
        mock_session = MagicMock()
        with patch('aiohttp.ClientSession', return_value=mock_session):
            client = LaMarzoccoCloudClient("test", "test", MOCK_SECRET_DATA)
            yield client

    async def test_handle_websocket_message_closing(self, mock_client) -> None:
        """Test handling websocket CLOSING message."""
        mock_ws = MagicMock()
        msg = MagicMock()
        msg.type = WSMsgType.CLOSING
        
        result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
            mock_ws, msg
        )
        
        assert result is True

    async def test_handle_websocket_message_closed(self, mock_client) -> None:
        """Test handling websocket CLOSED message."""
        mock_ws = MagicMock()
        msg = MagicMock()
        msg.type = WSMsgType.CLOSED
        
        result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
            mock_ws, msg
        )
        
        assert result is True

    async def test_handle_websocket_message_error(self, mock_client) -> None:
        """Test handling websocket ERROR message."""
        mock_ws = MagicMock()
        mock_ws.exception.return_value = Exception("Connection error")
        msg = MagicMock()
        msg.type = WSMsgType.ERROR
        
        result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
            mock_ws, msg
        )
        
        assert result is True

    async def test_handle_websocket_message_text_error(self, mock_client) -> None:
        """Test handling websocket text message with STOMP ERROR."""
        mock_ws = MagicMock()
        msg = MagicMock()
        msg.type = WSMsgType.TEXT
        
        # Create STOMP ERROR message
        error_msg = encode_stomp_ws_message(
            StompMessageType.ERROR,
            {"message": "Access denied"},
            "Authentication failed"
        )
        msg.data = error_msg
        
        result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
            mock_ws, msg
        )
        
        assert result is False

    async def test_handle_websocket_message_non_message_type(self, mock_client) -> None:
        """Test handling websocket text with non-MESSAGE STOMP type."""
        mock_ws = MagicMock()
        msg = MagicMock()
        msg.type = WSMsgType.TEXT
        
        # Create STOMP CONNECTED message
        connected_msg = encode_stomp_ws_message(
            StompMessageType.CONNECTED,
            {"version": "1.2"},
            None
        )
        msg.data = connected_msg
        
        result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
            mock_ws, msg
        )
        
        assert result is False

    async def test_handle_websocket_message_invalid_format(self, mock_client) -> None:
        """Test handling websocket message with invalid format."""
        mock_ws = MagicMock()
        msg = MagicMock()
        msg.type = WSMsgType.TEXT
        msg.data = "invalid message format"
        
        result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
            mock_ws, msg
        )
        
        assert result is False

    async def test_handle_websocket_message_with_valid_message(self, mock_client) -> None:
        """Test handling websocket message with valid STOMP MESSAGE."""
        mock_ws = MagicMock()
        msg = MagicMock()
        msg.type = WSMsgType.TEXT
        
        # Create simple valid STOMP MESSAGE
        simple_msg = encode_stomp_ws_message(
            StompMessageType.MESSAGE,
            {"destination": "/topic/test"},
            '{"test": "data"}'
        )
        msg.data = simple_msg
        
        # Mock the parse method to prevent actual parsing
        with patch.object(mock_client, '_LaMarzoccoCloudClient__parse_websocket_message') as mock_parse:
            result = await mock_client._LaMarzoccoCloudClient__handle_websocket_message(
                mock_ws, msg
            )
            
            mock_parse.assert_called_once_with('{"test": "data"}', None)
        
        assert result is False

    def test_parse_websocket_message_none(self, mock_client) -> None:
        """Test parsing None websocket message."""
        # Should not raise exception
        mock_client._LaMarzoccoCloudClient__parse_websocket_message(None, None)

    def test_parse_websocket_message_with_callback(self, mock_client) -> None:
        """Test parsing websocket message with notification callback."""
        config_json = '{"connected": true, "commands": [], "widgets": []}'
        callback = MagicMock()
        
        with patch('pylamarzocco.models.ThingDashboardWebsocketConfig.from_json') as mock_from_json:
            mock_config = MagicMock()
            mock_config.commands = []
            mock_from_json.return_value = mock_config
            
            mock_client._LaMarzoccoCloudClient__parse_websocket_message(config_json, callback)
            
            callback.assert_called_once_with(mock_config)

    def test_parse_websocket_message_pending_command(self, mock_client) -> None:
        """Test parsing websocket message with pending command."""
        command_id = "test-command-id"
        mock_future = MagicMock()
        mock_client._pending_commands[command_id] = mock_future
        
        config_json = '{"connected": true, "commands": [], "widgets": []}'
        
        with patch('pylamarzocco.models.ThingDashboardWebsocketConfig.from_json') as mock_from_json:
            mock_config = MagicMock()
            mock_command = MagicMock()
            mock_command.id = command_id
            mock_config.commands = [mock_command]
            mock_from_json.return_value = mock_config
            
            mock_client._LaMarzoccoCloudClient__parse_websocket_message(config_json, None)
            
            mock_future.set_result.assert_called_once_with(mock_command)

    async def test_websocket_setup_connection_basic(self, mock_client) -> None:
        """Test basic websocket setup connection."""
        mock_ws = MagicMock()
        mock_ws.send_str = AsyncMock()
        mock_ws.receive = AsyncMock()
        mock_ws.closed = False
        
        # Mock the CONNECTED response
        connected_response = MagicMock()
        connected_response.data = encode_stomp_ws_message(
            StompMessageType.CONNECTED,
            {"version": "1.2"},
            None
        )
        mock_ws.receive.return_value = connected_response
        
        with patch.object(mock_client, 'async_get_access_token', return_value="mock-token"):
            await mock_client._LaMarzoccoCloudClient__setup_websocket_connection(
                mock_ws, "MR123456"
            )
        
        # Verify CONNECT and SUBSCRIBE messages were sent
        assert mock_ws.send_str.call_count == 2
        
        # Verify websocket details were set
        assert mock_client.websocket is not None
        assert mock_client.websocket.connected

    async def test_websocket_setup_connection_authentication_failure(self, mock_client) -> None:
        """Test websocket setup when authentication fails."""
        mock_ws = MagicMock()
        mock_ws.send_str = AsyncMock()
        mock_ws.receive = AsyncMock()
        
        # Mock an ERROR response instead of CONNECTED
        error_response = MagicMock()
        error_response.data = encode_stomp_ws_message(
            StompMessageType.ERROR,
            {"message": "Authentication failed"},
            None
        )
        mock_ws.receive.return_value = error_response
        
        with patch.object(mock_client, 'async_get_access_token', return_value="mock-token"):
            with pytest.raises(Exception):  # Should raise ClientConnectionError
                await mock_client._LaMarzoccoCloudClient__setup_websocket_connection(
                    mock_ws, "MR123456"
                )

    async def test_websocket_connect_error_handling(self, mock_client) -> None:
        """Test websocket connection error handling."""
        with patch.object(mock_client._client, 'ws_connect') as mock_ws_connect:
            mock_ws_connect.side_effect = Exception("Connection failed")
            
            # Should not raise, should handle exception gracefully
            await mock_client.websocket_connect("MR123456", auto_reconnect=False)

    def test_websocket_details_disconnect_functionality(self) -> None:
        """Test WebSocketDetails disconnect functionality."""
        mock_ws = MagicMock()
        mock_ws.closed = False
        disconnect_callback = MagicMock()
        
        details = WebSocketDetails(ws=mock_ws, disconnect_callback=disconnect_callback)
        
        assert details.connected
        
        # Test disconnect without callback
        details_no_callback = WebSocketDetails(ws=mock_ws)
        assert details_no_callback.connected


"""Bluetooth class for La Marzocco machines."""

from __future__ import annotations

import asyncio
from functools import wraps
import json
import logging
from typing import Any, Callable, Concatenate, Coroutine

from bleak import BaseBleakScanner, BleakClient, BleakError, BleakScanner, BLEDevice
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak_retry_connector import BleakClientWithServiceCache, establish_connection

from pylamarzocco.const import (
    BluetoothReadSetting,
    BoilerType,
    MachineMode,
    SmartStandByType,
)
from pylamarzocco.exceptions import BluetoothConnectionFailed
from pylamarzocco.models import (
    BluetoothBoilerDetails,
    BluetoothCommandStatus,
    BluetoothMachineCapabilities,
    BluetoothSmartStandbyDetails,
)

_logger = logging.getLogger(__name__)

READ_CHARACTERISTIC = "0a0b7847-e12b-09a8-b04b-8e0922a9abab"
WRITE_CHARACTERISTIC = "0b0b7847-e12b-09a8-b04b-8e0922a9abab"
GET_TOKEN_CHARACTERISTIC = "0c0b7847-e12b-09a8-b04b-8e0922a9abab"
AUTH_CHARACTERISTIC = "0d0b7847-e12b-09a8-b04b-8e0922a9abab"

BT_MODEL_PREFIXES = ("MICRA", "MINI", "GS3")
IDLE_TIMEOUT = 30  # seconds


def disconnect_on_exception[
    T: "LaMarzoccoBluetoothClient", _R, **P
](
    func: Callable[Concatenate[T, P], Coroutine[Any, Any, _R]],
) -> Callable[Concatenate[T, P], Coroutine[Any, Any, _R]]:
    """Decorator to disconnect on exception."""

    @wraps(func)
    async def wrapper(
        self: T, *args: P.args, **kwargs: P.kwargs
    ) -> _R:
        try:
            return await func(self, *args, **kwargs)
        except (BleakError, TimeoutError, BluetoothConnectionFailed):
            # Disconnect on error (outside the lock to avoid deadlock)
            asyncio.create_task(self.disconnect())
            raise

    return wrapper


class LaMarzoccoBluetoothClient:
    """Class to interact with machine via Bluetooth."""

    def __init__(
        self,
        ble_device: BLEDevice,
        ble_token: str,
    ) -> None:
        """Initializes a new bluetooth client instance.
        
        Args:
            ble_device: The BLE device to connect to
            ble_token: Authentication token for the device
        """
        self._ble_token = ble_token
        self._address = ble_device.address
        self._ble_device = ble_device
        self._client: BleakClientWithServiceCache | None = None
        self._lock: asyncio.Lock = asyncio.Lock()
        self._disconnect_task: asyncio.Task[None] | None = None



    @property
    def is_connected(self) -> bool:
        """Return whether the client is currently connected."""
        return self._client is not None and self._client.is_connected

    async def _ensure_connected(self) -> None:
        """Ensure we're connected to the device, connecting if necessary."""
        async with self._lock:
            if self.is_connected:
                # Reset the disconnect timer
                self._reset_disconnect_timer()
                return
            
            _logger.debug("Connecting to Bluetooth device %s", self._address)
            try:
                self._client = await establish_connection(
                    BleakClientWithServiceCache,
                    self._ble_device,
                    self._ble_device.name or "Unknown",
                    max_attempts=3,
                )
                await self._authenticate()
            except (BleakError, TimeoutError, BluetoothConnectionFailed) as e:
                _logger.error("Failed to connect to Bluetooth device: %s", e)
                self._client = None
                raise
            else:
                _logger.debug("Successfully connected to Bluetooth device %s", self._address)
                # Start the disconnect timer
                self._reset_disconnect_timer()

    def _reset_disconnect_timer(self) -> None:
        """Reset the auto-disconnect timer."""
        # Cancel existing timer if any
        if self._disconnect_task is not None and not self._disconnect_task.done():
            self._disconnect_task.cancel()
        
        # Start new timer
        self._disconnect_task = asyncio.create_task(self._auto_disconnect())

    async def _auto_disconnect(self) -> None:
        """Automatically disconnect after idle timeout."""
        try:
            await asyncio.sleep(IDLE_TIMEOUT)
        except asyncio.CancelledError:
            # Timer was reset, this is normal
            pass
        else:
            _logger.debug("Auto-disconnect timer expired, disconnecting from %s", self._address)
            await self.disconnect()

    async def _disconnect_internal(self) -> None:
        """Internal disconnect that doesn't acquire lock (assumes lock is already held)."""
        # Cancel disconnect timer
        if self._disconnect_task is not None and not self._disconnect_task.done():
            self._disconnect_task.cancel()
            self._disconnect_task = None

        if self._client is not None and self._client.is_connected:
            _logger.debug("Disconnecting from Bluetooth device %s", self._address)
            try:
                await self._client.disconnect()
            except Exception as e:
                _logger.error("Error disconnecting from Bluetooth device: %s", e)
            finally:
                self._client = None

    async def disconnect(self) -> None:
        """Disconnect from the device."""
        async with self._lock:
            await self._disconnect_internal()

    @staticmethod
    async def discover_devices(
        scanner: BaseBleakScanner | BleakScanner | None = None,
    ) -> list[BLEDevice]:
        """Find machines based on model name."""
        ble_devices: list[BLEDevice] = []

        if scanner is None:
            scanner = BleakScanner()
        assert hasattr(scanner, "discover")
        devices: list[BLEDevice] = await scanner.discover()
        for device in devices:
            if device.name and device.name.startswith(BT_MODEL_PREFIXES):
                ble_devices.append(device)

        return ble_devices

    @staticmethod
    async def read_token(address_or_ble_device: BLEDevice | str) -> str:
        """Read the token from the machine.

        Only possible when machine is in pairing mode.
        """
        async with BleakClient(address_or_ble_device) as client:
            token = await client.read_gatt_char(GET_TOKEN_CHARACTERISTIC)
            return token.decode()

    @property
    def address(self) -> str:
        """Return the BT MAC address of the machine."""

        return self._address

    @disconnect_on_exception
    async def get_machine_mode(self) -> MachineMode:
        """Read the current machine mode"""
        return MachineMode(
            await self.__read_value_from_machine(BluetoothReadSetting.MACHINE_MODE)
        )

    @disconnect_on_exception
    async def get_machine_capabilities(self) -> BluetoothMachineCapabilities:
        """Get general machine information."""
        capabilities = await self.__read_value_from_machine(
            BluetoothReadSetting.MACHINE_CAPABILITIES
        )
        return BluetoothMachineCapabilities.from_dict(capabilities[0])

    @disconnect_on_exception
    async def get_tank_status(self) -> bool:
        """Get the current tank status."""
        return bool(
            await self.__read_value_from_machine(BluetoothReadSetting.TANK_STATUS)
        )

    @disconnect_on_exception
    async def get_boilers(self) -> list[BluetoothBoilerDetails]:
        """Get the boiler status."""
        boilers = await self.__read_value_from_machine(BluetoothReadSetting.BOILERS)
        return [BluetoothBoilerDetails.from_dict(boiler) for boiler in boilers]

    @disconnect_on_exception
    async def get_smart_standby_settings(self) -> BluetoothSmartStandbyDetails:
        """Get the smart standby settings."""
        data = await self.__read_value_from_machine(BluetoothReadSetting.SMART_STAND_BY)
        return BluetoothSmartStandbyDetails.from_dict(data)

    @disconnect_on_exception
    async def set_power(self, enabled: bool) -> BluetoothCommandStatus:
        """Power on the machine."""
        mode = "BrewingMode" if enabled else "StandBy"
        data = {
            "name": "MachineChangeMode",
            "parameter": {
                "mode": mode,
            },
        }
        await self.__write_bluetooth_json_message(data)
        return await self._check_command_status()

    @disconnect_on_exception
    async def set_steam(self, enabled: bool) -> BluetoothCommandStatus:
        """Enable or disable the steam boiler."""
        data = {
            "name": "SettingBoilerEnable",
            "parameter": {
                "identifier": "SteamBoiler",
                "state": enabled,
            },
        }
        await self.__write_bluetooth_json_message(data)
        return await self._check_command_status()

    @disconnect_on_exception
    async def set_smart_standby(
        self, enabled: bool, mode: SmartStandByType, minutes: int
    ) -> BluetoothCommandStatus:
        """Set the smart standby settings."""
        data = {
            "name": "SettingSmartStandby",
            "parameter": {"minutes": minutes, "mode": mode.value, "enabled": enabled},
        }
        await self.__write_bluetooth_json_message(data)
        return await self._check_command_status()

    @disconnect_on_exception
    async def set_temp(self, boiler: BoilerType, temperature: float) -> BluetoothCommandStatus:
        """Set boiler temperature (in Celsius)"""
        data = {
            "name": "SettingBoilerTarget",
            "parameter": {
                "identifier": boiler.value,
                "value": temperature,
            },
        }
        await self.__write_bluetooth_json_message(data)
        return await self._check_command_status()

    async def _authenticate(self) -> None:
        """Build authentication string and send it to the machine."""
        if self._client is None:
            raise BluetoothConnectionFailed("Client is not connected")
            
        auth_characteristic = await self._resolve_characteristic(AUTH_CHARACTERISTIC)

        try:
            await self._client.write_gatt_char(
                char_specifier=auth_characteristic,
                data=bytes(self._ble_token, "utf-8"),
                response=True,
            )
        except (BleakError, TimeoutError) as e:
            raise BluetoothConnectionFailed(
                f"Failed to connect to machine with Bluetooth: {e}"
            ) from e

    async def __read_value_from_machine(self, setting: BluetoothReadSetting) -> Any:
        await self.__write_bluetooth_message(setting.value, READ_CHARACTERISTIC)
        return json.loads(await self._read_bluetooth_message())

    async def _read_bluetooth_message(
        self, characteristic: str = READ_CHARACTERISTIC
    ) -> str:
        """Read a bluetooth message."""
        await self._ensure_connected()
        
        if self._client is None:
            raise BluetoothConnectionFailed("Client is not connected")

        read_characteristic = await self._resolve_characteristic(characteristic)
        result = await self._client.read_gatt_char(read_characteristic)
        return result.decode()
    
    async def _check_command_status(
        self,
        characteristic: str = WRITE_CHARACTERISTIC,
    ) -> BluetoothCommandStatus:
        """Check the status of a command sent via Bluetooth."""
        result = await self._read_bluetooth_message(characteristic)
        return BluetoothCommandStatus.from_json(result)

    async def __write_bluetooth_message(
        self,
        message: bytes | str,
        characteristic: str = WRITE_CHARACTERISTIC,
    ) -> None:
        """Connect to machine and write a message."""
        await self._ensure_connected()
        
        if self._client is None:
            raise BluetoothConnectionFailed("Client is not connected")

        # check if message is already bytes string
        if not isinstance(message, bytes):
            message = bytes(message, "utf-8")

        # append trailing zeros to message
        message += b"\x00"

        _logger.debug("Sending bluetooth message: %s to %s", message, characteristic)

        settings_characteristic = await self._resolve_characteristic(characteristic)

        await self._client.write_gatt_char(
            char_specifier=settings_characteristic,
            data=message,
            response=True,
        )

    async def __write_bluetooth_json_message(
        self,
        data: dict[str, Any],
        characteristic: str = WRITE_CHARACTERISTIC,
    ) -> None:
        """Write a json message to the machine."""

        await self.__write_bluetooth_message(
            characteristic=characteristic,
            message=json.dumps(data, separators=(",", ":")),
        )

    async def _resolve_characteristic(
        self, characteristic: str
    ) -> BleakGATTCharacteristic:
        """Resolve characteristic UUID from machine services."""
        if self._client is None:
            raise BluetoothConnectionFailed("Client is not connected")
            
        resolved_characteristic = self._client.services.get_characteristic(
            characteristic
        )
        if resolved_characteristic is not None:
            return resolved_characteristic

        _logger.debug(
            "Characteristic %s not found in cache, clearing cache and retrying.",
            characteristic,
        )
        await self._client.clear_cache()

        resolved_characteristic = self._client.services.get_characteristic(
            characteristic
        )
        if resolved_characteristic is not None:
            return resolved_characteristic

        # Can't resolve characteristic - clear cache and schedule disconnect
        _logger.info(
            "Could not find characteristic %s on machine. Clearing cache and disconnecting.",
            characteristic,
        )
        await self._client.clear_cache()
        # Schedule disconnect outside the lock to avoid deadlock
        asyncio.create_task(self.disconnect())
        raise BluetoothConnectionFailed(
            f"Could not find characteristic {characteristic} on machine."
        )



"""Authentication related utilities."""


import base64
import hashlib
import time
import uuid
from dataclasses import dataclass, field

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.ec import (
    ECDSA,
    SECP256R1,
    EllipticCurvePrivateKey,
    generate_private_key,
)
from mashumaro import field_options
from mashumaro.mixins.json import DataClassJSONMixin


def b64(data: bytes) -> str:
    """Base64 encode bytes to ASCII string."""
    return base64.b64encode(data).decode("ascii")


@dataclass
class InstallationKey(DataClassJSONMixin):
    """Holds key material derived from installation ID."""

    secret: bytes = field(
        metadata=field_options(serialize=b64, deserialize=base64.b64decode)
    )
    private_key: EllipticCurvePrivateKey = field(
        metadata=field_options(
            serialize=lambda key: b64(
                key.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            ),
            deserialize=lambda s: serialization.load_der_private_key(
                base64.b64decode(s), password=None
            ),
        )
    )
    installation_id: str

    @property
    def public_key_b64(self) -> str:
        """Return public key in base64-encoded DER format."""
        pub_bytes = self.private_key.public_key().public_bytes(
            serialization.Encoding.DER, serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return b64(pub_bytes)

    @property
    def base_string(self) -> str:
        """Return base string: installation_id.sha256(public_key_der_bytes)"""
        pub_bytes = self.private_key.public_key().public_bytes(
            serialization.Encoding.DER, serialization.PublicFormat.SubjectPublicKeyInfo
        )
        pub_hash_b64 = b64(
            hashlib.sha256(pub_bytes).digest()
        )  # Hash raw bytes, not base64 string
        return f"{self.installation_id}.{pub_hash_b64}"

def generate_request_proof(base_string: str, secret32: bytes) -> str:
    """La Marzocco's custom proof generation algorithm (Y5.e equivalent)"""
    if len(secret32) != 32:
        raise ValueError("secret must be 32 bytes")

    work = bytearray(secret32)  # Make mutable copy

    for byte_val in base_string.encode("utf-8"):
        idx = byte_val % 32
        shift_idx = (idx + 1) % 32
        shift_amount = work[shift_idx] & 7  # 0-7 bit shift

        # XOR then rotate left
        xor_result = byte_val ^ work[idx]
        rotated = (
            (xor_result << shift_amount) | (xor_result >> (8 - shift_amount))
        ) & 0xFF
        work[idx] = rotated

    return b64(hashlib.sha256(work).digest())


def generate_extra_request_headers(installation_key: InstallationKey) -> dict[str, str]:
    """Generate extra headers for normal API calls after authentication"""

    # Generate nonce and timestamp
    nonce = str(uuid.uuid4()).lower()
    timestamp = str(int(time.time() * 1000))  # milliseconds

    # Create proof using Y5.e algorithm: installation_id.nonce.timestamp
    proof_input = f"{installation_key.installation_id}.{nonce}.{timestamp}"
    proof = generate_request_proof(proof_input, installation_key.secret)

    # Create signature data: installation_id.nonce.timestamp.proof
    signature_data = f"{proof_input}.{proof}"

    # Sign with ECDSA
    signature = installation_key.private_key.sign(
        signature_data.encode("utf-8"), ECDSA(hashes.SHA256())
    )
    signature_b64 = b64(signature)

    # Return headers
    return {
        "X-App-Installation-Id": installation_key.installation_id,
        "X-Timestamp": timestamp,
        "X-Nonce": nonce,
        "X-Request-Signature": signature_b64,
    }


def generate_installation_key(installation_id: str) -> InstallationKey:
    """Generate the key material from installation ID."""

    def derive_secret_bytes(installation_id: str, pub_der_bytes: bytes) -> bytes:
        pub_b64 = b64(pub_der_bytes)
        inst_hash = hashlib.sha256(installation_id.encode("utf-8")).digest()
        inst_hash_b64 = b64(inst_hash)
        triple = f"{installation_id}.{pub_b64}.{inst_hash_b64}"
        return hashlib.sha256(triple.encode("utf-8")).digest()  # 32 bytes

    private_key = generate_private_key(SECP256R1())

    pub_bytes = private_key.public_key().public_bytes(
        serialization.Encoding.DER, serialization.PublicFormat.SubjectPublicKeyInfo
    )

    secret_bytes = derive_secret_bytes(installation_id, pub_bytes)
    return InstallationKey(
        installation_id=installation_id, secret=secret_bytes, private_key=private_key
    )



"""WebSocket utility functions."""

from pylamarzocco.const import StompMessageType

def encode_stomp_ws_message(
    msg_type: StompMessageType, headers: dict[str, str], body: str | None = None
) -> str:
    """Encode STOMP WebSocket message."""
    fragments: list[str] = []
    fragments.append(str(msg_type))
    for key, value in headers.items():
        fragments.append(f"{key}:{value}")
    msg = "\n".join(fragments)
    msg += "\n\n"
    if body:
        msg += body
    msg += "\x00"
    return msg


def decode_stomp_ws_message(
    msg: str,
) -> tuple[StompMessageType, dict[str, str], str | None]:
    """Decode STOMP WebSocket message."""
    header, data = msg.split("\n\n", 1)
    headers: dict[str, str] = {}
    metadata = header.split("\n")
    msg_type = StompMessageType(metadata[0])
    for header in metadata[1:]:
        key, value = header.split(":", 1)
        headers[key] = value
    if data.endswith("\x00"):
        data = data[:-1]
    return msg_type, headers, data





"""Tests for websocket related functionality."""

import json

from syrupy import SnapshotAssertion

from pylamarzocco.const import StompMessageType
from pylamarzocco.models import ThingDashboardWebsocketConfig
from pylamarzocco.util import decode_stomp_ws_message, encode_stomp_ws_message

from .conftest import load_fixture


async def test_encode_stomp_ws_message(snapshot: SnapshotAssertion) -> None:
    """Test the encode_stomp_ws_message function."""

    fixture = load_fixture("machine", "config_micra.json")
    device_config = ThingDashboardWebsocketConfig.from_dict(fixture)
    headers = {
        "destination": "/topic/test",
        "id": "test-id",
        "subscription": "test-subscription",
    }
    message = encode_stomp_ws_message(
        StompMessageType.MESSAGE, headers, json.dumps(fixture)
    )
    assert message == snapshot

    msg_type, headers_decoded, body = decode_stomp_ws_message(message)

    assert body
    assert msg_type is StompMessageType.MESSAGE
    assert headers_decoded == headers
    assert ThingDashboardWebsocketConfig.from_json(body) == device_config



"""Models for authentication."""

from dataclasses import dataclass, field
from time import time

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.json import DataClassJSONMixin

TOKEN_EXPIRATION = 60 * 60 # 1 hour

@dataclass(kw_only=True, frozen=True)
class AccessToken(DataClassJSONMixin):
    """Base for access token response model."""

    access_token: str = field(metadata=field_options(alias="accessToken"))
    refresh_token: str = field(metadata=field_options(alias="refreshToken"))
    expires_at: float = field(default_factory=lambda: time() + TOKEN_EXPIRATION)

@dataclass(kw_only=True, frozen=True)
class SigninTokenRequest(DataClassJSONMixin):
    """Token request model."""

    username: str
    password: str

@dataclass(kw_only=True, frozen=True)
class RefreshTokenRequest(DataClassJSONMixin):
    """Token request model."""

    username: str
    refresh_token: str = field(metadata=field_options(alias="refreshToken"))
    class Config(BaseConfig):
        """Config for Mashumaro serialization."""
        serialize_by_alias = True


"""Constants for La Marzocco Cloud."""

from __future__ import annotations

from enum import StrEnum

BASE_URL = "lion.lamarzocco.io"
CUSTOMER_APP_URL = f"https://{BASE_URL}/api/customer-app"


class MachineMode(StrEnum):
    """Machine states."""

    BREWING_MODE = "BrewingMode"
    ECO_MODE = "EcoMode"
    STANDBY = "StandBy"


class MachineState(StrEnum):
    """Machine statuses."""

    STANDBY = "StandBy"
    POWERED_ON = "PoweredOn"
    BREWING = "Brewing"
    OFF = "Off"


class PreExtractionMode(StrEnum):
    """Pre-extraction modes."""

    PREINFUSION = "PreInfusion"
    PREBREWING = "PreBrewing"
    DISABLED = "Disabled"


class WidgetType(StrEnum):
    """Widget types."""

    CM_MACHINE_STATUS = "CMMachineStatus"
    CM_COFFEE_BOILER = "CMCoffeeBoiler"
    CM_STEAM_BOILER_LEVEL = "CMSteamBoilerLevel"
    CM_PRE_EXTRACTION = "CMPreExtraction"
    CM_PRE_BREWING = "CMPreBrewing"
    CM_BACK_FLUSH = "CMBackFlush"
    CM_MACHINE_GROUP_STATUS = "CMMachineGroupStatus"
    CM_MACHINE_MODULE_STATUS = "CMMachineModuleStatus"
    CM_STEAM_BOILER_TEMPERATURE = "CMSteamBoilerTemperature"
    CM_GROUP_DOSES = "CMGroupDoses"
    CM_PRE_INFUSION_ENABLE = "CMPreInfusionEnable"
    CM_PRE_INFUSION = "CMPreInfusion"
    CM_BREW_BY_WEIGHT_DOSES = "CMBrewByWeightDoses"
    CM_CUP_WARMER = "CMCupWarmer"
    CM_HOT_WATER_DOSE = "CMHotWaterDose"
    CM_AUTO_FLUSH = "CMAutoFlush"
    CM_RINSE_FLUSH = "CMRinseFlush"
    CM_STEAM_FLUSH = "CMSteamFlush"
    CM_NO_WATER = "CMNoWater"
    CM_MODULE_AND_TAPS_TEMPERATURE = "CMModuleAndTapsTemperature"
    G_MACHINE_STATUS = "GMachineStatus"
    G_DOSES = "GDoses"
    G_SINGLE_DOSE_MODE = "GSingleDoseMode"
    G_BARISTA_LIGHT = "GBaristaLight"
    G_HOPPER_OPENED = "GHopperOpened"
    G_MIRROR_DOSES = "GMirrorDoses"
    G_MORE_DOSE = "GMoreDose"
    G_GRIND_WITH = "GGrindWith"
    G_SPEED = "GSpeed"
    # scale
    THING_SCALE = "ThingScale"
    # statistic widgets
    COFFEE_AND_FLUSH_TREND = "COFFEE_AND_FLUSH_TREND"
    LAST_COFFEE = "LAST_COFFEE"
    COFFEE_AND_FLUSH_COUNTER = "COFFEE_AND_FLUSH_COUNTER"


class CommandStatus(StrEnum):
    """Command statuses."""

    SUCCESS = "Success"
    ERROR = "Error"
    TIMEOUT = "Timeout"
    PENDING = "Pending"
    IN_PROGRESS = "InProgress"


class SteamTargetLevel(StrEnum):
    """Steam target levels."""

    LEVEL_1 = "Level1"
    LEVEL_2 = "Level2"
    LEVEL_3 = "Level3"


class StompMessageType(StrEnum):
    """Stomp message types."""

    CONNECT = "CONNECT"
    CONNECTED = "CONNECTED"
    SUBSCRIBE = "SUBSCRIBE"
    UNSUBSCRIBE = "UNSUBSCRIBE"
    MESSAGE = "MESSAGE"
    ERROR = "ERROR"


class DeviceType(StrEnum):
    """Device types."""

    MACHINE = "CoffeeMachine"
    GRINDER = "Grinder"


class ModelCode(StrEnum):
    """Model codes."""

    LINEA_MINI = "LINEAMINI"
    LINEA_MICRA = "LINEAMICRA"
    LINEA_MINI_R = "LINEAMINIR"
    GS3 = "GS3"
    GS3_MP = "GS3MP"
    GS3_AV = "GS3AV"
    PICO_GRINDER = "PICOGRINDER"
    SWAN_GRINDER = "SWANGRINDER"


class ModelName(StrEnum):
    """Model names."""

    LINEA_MINI = "Linea Mini"
    LINEA_MICRA = "Linea Micra"
    LINEA_MINI_R = "Linea Mini R"
    GS3 = "GS3"
    GS3_MP = "GS3 MP"
    GS3_AV = "GS3 AV"
    PICO_GRINDER = "Pico"
    SWAN_GRINDER = "Swan"

    @classmethod
    def from_string(cls, name: str) -> ModelName:
        """Convert a string to a ModelName enum."""
        mapping = {
            "GS3MP": "GS3 MP",
            "GS3AV": "GS3 AV",
            "LINEAMINI2023": "Linea Mini R",
            "LINEAMICRA": "Linea Micra",
            "LINEAMINI": "Linea Mini",
            "MICRA": "Linea Micra",
            "PICOGRINDER": "Pico",
            "SWANGRINDER": "Swan",
        }
        if (key := "".join(name.upper().split())) not in mapping:
            raise ValueError(f"Invalid model name: {name}")
        return cls(mapping[key])


class FirmwareType(StrEnum):
    """Firmware types."""

    MACHINE = "Machine"
    GATEWAY = "Gateway"


class DoseIndexType(StrEnum):
    """Dose index types."""

    BY_GROUP = "ByGroup"
    BY_DOSE = "ByDose"  # TODO: Check if this is correct


class DoseMode(StrEnum):
    """Dose modes."""

    CONTINUOUS = "Continuous"
    PULSES_TYPE = "PulsesType"
    DOSE_1 = "Dose1"
    DOSE_2 = "Dose2"


class DoseIndex(StrEnum):
    """Dose index types."""

    CONTINUOUS = "Continuous"
    BY_GROUP = "ByGroup"
    DOSE_A = "DoseA"
    DOSE_B = "DoseB"
    DOSE_C = "DoseC"
    DOSE_D = "DoseD"


class SmartStandByType(StrEnum):
    """Smart Standby types."""

    LAST_BREW = "LastBrewing"
    POWER_ON = "PowerOn"


class BoilerStatus(StrEnum):
    """Boiler statuses."""

    STAND_BY = "StandBy"
    HEATING = "HeatingUp"
    READY = "Ready"
    NO_WATER = "NoWater"
    OFF = "Off"


class WeekDay(StrEnum):
    """Week days."""

    MONDAY = "Monday"
    TUESDAY = "Tuesday"
    WEDNESDAY = "Wednesday"
    THURSDAY = "Thursday"
    FRIDAY = "Friday"
    SATURDAY = "Saturday"
    SUNDAY = "Sunday"


class UpdateStatus(StrEnum):
    """Update statuses."""

    TO_UPDATE = "ToUpdate"
    PENDING = "Pending"
    IN_PROGRESS = "InProgress"
    UPDATED = "Updated"


class UpdateProgressInfo(StrEnum):
    """Update progress info."""

    DOWNLOAD = "download"
    REBOOTING = "rebooting"
    STARTING_PROCESS = "starting process"


class BoilerType(StrEnum):
    """La Marzocco Coffee Machine Boilers."""

    COFFEE = "CoffeeBoiler1"
    STEAM = "SteamBoiler"


class BackFlushStatus(StrEnum):
    """Back flush statuses."""

    REQUESTED = "Requested"
    CLEANING = "Cleaning"
    OFF = "Off"


class BluetoothReadSetting(StrEnum):
    """Declare what to read from the Bluetooth device."""

    MACHINE_CAPABILITIES = "machineCapabilities"
    MACHINE_MODE = "machineMode"
    TANK_STATUS = "tankStatus"
    BOILERS = "boilers"
    SMART_STAND_BY = "smartStandBy"



"""La Marzocco Cloud API Client."""

from __future__ import annotations

import asyncio
import logging
import time
import uuid
from asyncio import Future, wait_for
from collections.abc import Callable
from http import HTTPMethod
from typing import Any

from aiohttp import (
    ClientConnectionError,
    ClientSession,
    ClientTimeout,
    ClientWebSocketResponse,
    ClientWSTimeout,
    WSMessage,
    WSMsgType,
)
from aiohttp.client_exceptions import ClientError, InvalidURL

from pylamarzocco.const import (
    BASE_URL,
    CUSTOMER_APP_URL,
    CommandStatus,
    PreExtractionMode,
    SmartStandByType,
    SteamTargetLevel,
    StompMessageType,
    WidgetType,
)
from pylamarzocco.exceptions import AuthFail, RequestNotSuccessful
from pylamarzocco.models import (
    AccessToken,
    CoffeeAndFlushCounter,
    CoffeeAndFlushTrend,
    CommandResponse,
    LastCoffeeList,
    PrebrewSettingTimes,
    RefreshTokenRequest,
    SigninTokenRequest,
    Thing,
    ThingDashboardConfig,
    ThingDashboardWebsocketConfig,
    ThingSchedulingSettings,
    ThingSettings,
    ThingStatistics,
    UpdateDetails,
    WakeUpScheduleSettings,
    WebSocketDetails,
)
from pylamarzocco.util import (
    InstallationKey,
    decode_stomp_ws_message,
    encode_stomp_ws_message,
    generate_extra_request_headers,
    generate_request_proof,
    is_success,
)

_LOGGER = logging.getLogger(__name__)


TOKEN_TIME_TO_REFRESH = 10 * 60  # 10 minutes before expiration
PENDING_COMMAND_TIMEOUT = 10


class LaMarzoccoCloudClient:
    """La Marzocco Cloud Client."""

    _client: ClientSession

    def __init__(
        self,
        username: str,
        password: str,
        installation_key: InstallationKey,
        client: ClientSession | None = None,
    ) -> None:
        """Set the cloud client up."""
        self._client = ClientSession() if client is None else client
        self._username = username
        self._password = password
        self._installation_key = installation_key
        self._access_token: AccessToken | None = None
        self._access_token_lock = asyncio.Lock()
        self._pending_commands: dict[str, Future[CommandResponse]] = {}
        self.websocket = WebSocketDetails()

    # region Authentication
    async def async_register_client(self) -> None:
        """Register a new client."""

        headers = {
            "X-App-Installation-Id": self._installation_key.installation_id,
            "X-Request-Proof": generate_request_proof(
                self._installation_key.base_string, self._installation_key.secret
            ),
        }
        body = {
            "pk": self._installation_key.public_key_b64,
        }
        try:
            response = await self._client.post(
                url=f"{CUSTOMER_APP_URL}/auth/init",
                headers=headers,
                json=body,
            )
        except ClientError as ex:
            raise RequestNotSuccessful(
                "Error during HTTP request."
                + f"Request auth to endpoint failed with error: {ex}"
            ) from ex

        if is_success(response):
            _LOGGER.info("Registration successful.")
            return

        if response.status == 401:
            raise AuthFail("Invalid username or password")

        raise RequestNotSuccessful(
            f"Request to auth endpoint failed with status code {response.status}"
            + f"response: {await response.text()}"
        )

    async def async_get_access_token(self) -> str:
        """Return a valid access token."""
        async with self._access_token_lock:
            if (
                self._access_token is None
                or self._access_token.expires_at < time.time()
            ):
                self._access_token = await self._async_sign_in()

            elif self._access_token.expires_at < time.time() + TOKEN_TIME_TO_REFRESH:
                self._access_token = await self._async_refresh_token()

            return self._access_token.access_token

    async def _async_sign_in(self) -> AccessToken:
        """Get a new access token."""
        _LOGGER.debug("Getting new access token")
        return await self.__async_get_token(
            f"{CUSTOMER_APP_URL}/auth/signin",
            SigninTokenRequest(
                username=self._username, password=self._password
            ).to_dict(),
        )

    async def _async_refresh_token(self) -> AccessToken:
        """Refresh a access token."""
        if not self._access_token:
            raise ValueError("No access token available")
        _LOGGER.debug("Refreshing access token")
        return await self.__async_get_token(
            f"{CUSTOMER_APP_URL}/auth/refreshtoken",
            RefreshTokenRequest(
                username=self._username, refresh_token=self._access_token.refresh_token
            ).to_dict(),
        )

    async def __async_get_token(self, url: str, data: dict[str, Any]) -> AccessToken:
        """Wrapper for a token request."""
        try:
            response = await self._client.post(
                url=url,
                headers=generate_extra_request_headers(self._installation_key),
                json=data,
            )
        except ClientError as ex:
            raise RequestNotSuccessful(
                "Error during HTTP request."
                + f"Request auth to endpoint failed with error: {ex}"
            ) from ex
        if is_success(response):
            json_response = await response.json()
            return AccessToken.from_dict(json_response)

        if response.status == 401:
            raise AuthFail("Invalid username or password")

        raise RequestNotSuccessful(
            f"Request t auth endpoint failed with status code {response.status}"
            + f"response: {await response.text()}"
        )

    # endregion

    async def _rest_api_call(
        self,
        url: str,
        method: HTTPMethod,
        data: dict[str, Any] | None = None,
        timeout: int = 10,
    ) -> dict:
        """Wrapper for the API call."""

        access_token = await self.async_get_access_token()
        headers = {
            **generate_extra_request_headers(self._installation_key),
            "Authorization": f"Bearer {access_token}",
        }

        try:
            response = await self._client.request(
                method=method,
                url=url,
                json=data,
                timeout=ClientTimeout(total=timeout),
                headers=headers,
            )
        except ClientError as ex:
            raise RequestNotSuccessful(
                f"Error during HTTP request. Request to endpoint {url} failed with error: {ex}"
            ) from ex

        # ensure status code indicates success
        if is_success(response):
            json_response = await response.json()
            _LOGGER.debug("Request to %s successful", url)
            _LOGGER.debug("Response: %s", json_response)
            return json_response

        if response.status == 401:
            raise AuthFail("Authentication failed.")

        raise RequestNotSuccessful(
            f"Request to endpoint {response.url} failed with status code {response.status}.\n"
            + f"Full response: {await response.text()}"
        )

    # region config
    async def list_things(self) -> list[Thing]:
        """Get all things."""
        url = f"{CUSTOMER_APP_URL}/things"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        return [Thing.from_dict(device) for device in result]

    async def get_thing_dashboard(self, serial_number: str) -> ThingDashboardConfig:
        """Get the dashboard of a thing."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/dashboard"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        return ThingDashboardConfig.from_dict(result)

    async def get_thing_settings(self, serial_number: str) -> ThingSettings:
        """Get the settings of a thing."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/settings"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        return ThingSettings.from_dict(result)

    async def get_thing_statistics(self, serial_number: str) -> ThingStatistics:
        """Get the statistics of a thing."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/stats"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        return ThingStatistics.from_dict(result)

    async def get_thing_firmware(self, serial_number: str) -> UpdateDetails:
        """Get the firmware settings of a thing."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/update-fw"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        return UpdateDetails.from_dict(result)

    async def get_thing_schedule(self, serial_number: str) -> ThingSchedulingSettings:
        """Get the schedule of a thing."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/scheduling"
        result = await self._rest_api_call(url=url, method=HTTPMethod.GET)
        return ThingSchedulingSettings.from_dict(result)

    async def _get_thing_extended_statistics(
        self, serial_number: str, widget: WidgetType, **kwargs: Any
    ) -> dict:
        """Get the extended statistics of a thing."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/stats/{widget}/1"

        # Append query parameters if kwargs is provided
        if kwargs:
            query_params = "&".join(f"{key}={value}" for key, value in kwargs.items())
            url = f"{url}?{query_params}"

        result = await self._rest_api_call(url=url, method=HTTPMethod.GET, timeout=10)
        return result["output"]

    async def get_thing_coffee_and_flush_trend(
        self, serial_number: str, days: int, timezone: str
    ) -> CoffeeAndFlushTrend:
        """Get the last coffee and flush trend of a thing."""
        result = await self._get_thing_extended_statistics(
            serial_number=serial_number,
            widget=WidgetType.COFFEE_AND_FLUSH_TREND,
            days=days,
            timezone=timezone,
        )
        return CoffeeAndFlushTrend.from_dict(result)

    async def get_thing_last_coffee(
        self, serial_number: str, days: int
    ) -> LastCoffeeList:
        """Get the last coffee of a thing."""
        result = await self._get_thing_extended_statistics(
            serial_number=serial_number,
            widget=WidgetType.LAST_COFFEE,
            days=days,
        )
        return LastCoffeeList.from_dict(result)

    async def get_thing_coffee_and_flush_counter(
        self, serial_number: str
    ) -> CoffeeAndFlushCounter:
        """Get the coffee and flush counter of a thing."""
        result = await self._get_thing_extended_statistics(
            serial_number=serial_number,
            widget=WidgetType.COFFEE_AND_FLUSH_COUNTER,
        )
        return CoffeeAndFlushCounter.from_dict(result)

    # endregion

    # region websocket
    async def websocket_connect(
        self,
        serial_number: str,
        notification_callback: Callable[[ThingDashboardWebsocketConfig], Any]
        | None = None,
        connect_callback: Callable[[], Any] | None = None,
        disconnect_callback: Callable[[], Any] | None = None,
        auto_reconnect: bool = True,
    ) -> None:
        """Connect to the websocket of the machine."""
        while auto_reconnect:
            try:
                async with await self._client.ws_connect(
                    f"wss://{BASE_URL}/ws/connect",
                    timeout=ClientWSTimeout(ws_receive=None, ws_close=10.0),
                    heartbeat=15,
                    headers=generate_extra_request_headers(self._installation_key),
                ) as ws:
                    try:
                        await self.__setup_websocket_connection(ws, serial_number)
                        if connect_callback is not None:
                            connect_callback()
                        async for msg in ws:
                            if await self.__handle_websocket_message(
                                ws, msg, notification_callback
                            ):
                                break
                    except asyncio.CancelledError:
                        _LOGGER.debug("WebSocket cancellation requested")
                        await self.websocket.disconnect()
                        raise
            except InvalidURL:
                _LOGGER.error("Invalid URL for websocket.")
                auto_reconnect = False
            except TimeoutError:
                _LOGGER.debug("Websocket disconnected: Connection timed out")
            except ClientError as err:
                _LOGGER.debug("Websocket disconnected: Could not connect: %s", err)
                auto_reconnect = False
            except asyncio.CancelledError:
                _LOGGER.debug("WebSocket cancellation successful")
                auto_reconnect = False
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Websocket disconnected with error")
                auto_reconnect = False
            finally:
                if disconnect_callback is not None:
                    disconnect_callback()

    async def __setup_websocket_connection(
        self,
        ws: ClientWebSocketResponse,
        serial_number: str,
    ) -> None:
        """Setup the websocket connection."""

        connect_msg = encode_stomp_ws_message(
            StompMessageType.CONNECT,
            {
                "host": BASE_URL,
                "accept-version": "1.2,1.1,1.0",
                "heart-beat": "0,0",
                "Authorization": f"Bearer {await self.async_get_access_token()}",
            },
        )
        _LOGGER.debug("Connecting to websocket.")
        await ws.send_str(connect_msg)

        msg = await ws.receive()
        _LOGGER.debug("Received websocket message: %s", msg.data)
        result, _, _ = decode_stomp_ws_message(str(msg.data))
        if result is not StompMessageType.CONNECTED:
            raise ClientConnectionError("No connected message")

        _LOGGER.debug("Subscribing to websocket.")
        subscription_id = str(uuid.uuid4())
        subscribe_msg = encode_stomp_ws_message(
            StompMessageType.SUBSCRIBE,
            {
                "destination": f"/ws/sn/{serial_number}/dashboard",
                "ack": "auto",
                "id": subscription_id,
                "content-length": "0",
            },
        )
        await ws.send_str(subscribe_msg)

        async def disconnect_websocket() -> None:
            _LOGGER.debug("Disconnecting websocket")
            if ws.closed:
                return
            disconnect_msg = encode_stomp_ws_message(
                StompMessageType.UNSUBSCRIBE,
                {
                    "id": subscription_id,
                },
            )
            await ws.send_str(disconnect_msg)
            await ws.close()

        self.websocket = WebSocketDetails(ws, disconnect_websocket)

    async def __handle_websocket_message(
        self,
        ws: ClientWebSocketResponse,
        msg: WSMessage,
        notification_callback: Callable[[ThingDashboardWebsocketConfig], Any]
        | None = None,
    ) -> bool:
        """Handle receiving a websocket message. Return True for disconnect."""
        if msg.type in (WSMsgType.CLOSING, WSMsgType.CLOSED):
            _LOGGER.debug("Websocket disconnected gracefully")
            return True
        if msg.type == WSMsgType.ERROR:
            _LOGGER.warning("Websocket disconnected with error %s", ws.exception())
            return True
        _LOGGER.debug("Received websocket message: %s", msg)
        try:
            msg_type, _, data = decode_stomp_ws_message(str(msg.data))
            if msg_type is StompMessageType.ERROR:
                _LOGGER.warning("Websocket error message: %s", data)
            elif msg_type is not StompMessageType.MESSAGE:
                _LOGGER.warning("Non MESSAGE-type message: %s", msg.data)
            else:
                self.__parse_websocket_message(data, notification_callback)
        except ValueError as ex:
            _LOGGER.warning(
                "Error parsing websocket message: %s. Message was: %s", ex, msg.data
            )
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.exception("Error during callback: %s", ex)
        return False

    def __parse_websocket_message(
        self,
        message: str | None,
        notification_callback: Callable[[ThingDashboardWebsocketConfig], Any]
        | None = None,
    ) -> None:
        """Parse the websocket message."""
        if message is None:
            return
        config = ThingDashboardWebsocketConfig.from_json(message)

        # notify if there is the result for a pending command
        for command in config.commands:
            if command.id in self._pending_commands:
                self._pending_commands[command.id].set_result(command)

        # notify any external listeners
        if notification_callback is not None:
            notification_callback(config)

    # endregion
    # region commands

    async def __execute_command(
        self, serial_number: str, command: str, data: dict[str, Any] | None = None
    ) -> bool:
        """Execute a command on a machine."""
        response = await self._rest_api_call(
            url=f"{CUSTOMER_APP_URL}/things/{serial_number}/command/{command}",
            method=HTTPMethod.POST,
            data=data,
        )
        cr = CommandResponse.from_dict(response[0])
        future: Future[CommandResponse] = Future()
        self._pending_commands[cr.id] = future

        # if the websocket is closed we don't want to wait for confirmation
        if not self.websocket.connected:
            return True

        try:
            # Wait for the future to be completed or timeout
            pending_result = await wait_for(future, PENDING_COMMAND_TIMEOUT)
        except TimeoutError:
            _LOGGER.debug("Timed out waiting for websocket condition")
            self._pending_commands.pop(cr.id, None)
            return False

        # Clean up the future if it's still in the dictionary
        self._pending_commands.pop(cr.id, None)

        if pending_result.status is CommandStatus.SUCCESS:
            return True
        _LOGGER.debug(
            "Command to %s failed with status %s, error_details: %s",
            command,
            pending_result.status,
            pending_result.error_code or "",
        )
        return False

    async def set_power(
        self,
        serial_number: str,
        enabled: bool,
    ) -> bool:
        """Turn power of machine on or off"""

        mode = "BrewingMode" if enabled else "StandBy"

        data = {"mode": mode}
        return await self.__execute_command(
            serial_number, "CoffeeMachineChangeMode", data
        )

    async def set_steam(
        self,
        serial_number: str,
        enabled: bool,
        boiler_index: int = 1,
    ) -> bool:
        """Turn steam boiler on or off"""

        data = {
            "boilerIndex": boiler_index,
            "enabled": enabled,
        }
        return await self.__execute_command(
            serial_number, "CoffeeMachineSettingSteamBoilerEnabled", data
        )

    async def set_steam_target_level(
        self,
        serial_number: str,
        target_level: SteamTargetLevel,
        boiler_index: int = 1,
    ) -> bool:
        """Set steam boiler target level"""

        data = {
            "boilerIndex": boiler_index,
            "targetLevel": target_level.value,
        }
        return await self.__execute_command(
            serial_number, "CoffeeMachineSettingSteamBoilerTargetLevel", data
        )

    async def set_coffee_target_temperature(
        self, serial_number: str, target_temperature: float, boiler_index: int = 1
    ) -> bool:
        """Set the target temperature for the coffee boiler."""
        data = {
            "boilerIndex": boiler_index,
            "targetTemperature": round(target_temperature, 1),
        }
        return await self.__execute_command(
            serial_number, "CoffeeMachineSettingCoffeeBoilerTargetTemperature", data
        )

    async def start_backflush_cleaning(
        self,
        serial_number: str,
    ) -> bool:
        """Start backflush cleaning"""

        data = {"enabled": True}
        return await self.__execute_command(
            serial_number, "CoffeeMachineBackFlushStartCleaning", data
        )

    async def change_pre_extraction_mode(
        self, serial_number: str, prebrew_mode: PreExtractionMode
    ) -> bool:
        """Change pre-extraction mode"""

        data = {
            "mode": prebrew_mode.value,
        }
        return await self.__execute_command(
            serial_number, "CoffeeMachinePreBrewingChangeMode", data
        )

    async def change_pre_extraction_times(
        self,
        serial_number: str,
        times: PrebrewSettingTimes,
    ) -> bool:
        """Change pre-extraction times"""

        return await self.__execute_command(
            serial_number, "CoffeeMachinePreBrewingSettingTimes", times.to_dict()
        )

    async def set_smart_standby(
        self, serial_number: str, enabled: bool, minutes: int, after: SmartStandByType
    ) -> bool:
        """Set smart standby"""

        data = {"enabled": enabled, "minutes": minutes, "after": after.value}
        return await self.__execute_command(
            serial_number, "CoffeeMachineSettingSmartStandBy", data
        )

    async def delete_wakeup_schedule(
        self,
        serial_number: str,
        schedule_id: str,
    ) -> bool:
        """Delete a smart wakeup schedule"""
        data = {"id": schedule_id}
        return await self.__execute_command(
            serial_number, "CoffeeMachineDeleteWakeUpSchedule", data
        )

    async def set_wakeup_schedule(
        self,
        serial_number: str,
        schedule: WakeUpScheduleSettings,
    ) -> bool:
        """Set smart wakeup schedule"""
        return await self.__execute_command(
            serial_number, "CoffeeMachineSetWakeUpSchedule", schedule.to_dict()
        )

    async def update_firmware(
        self,
        serial_number: str,
    ) -> UpdateDetails:
        """Install firmware update."""
        url = f"{CUSTOMER_APP_URL}/things/{serial_number}/update-fw"
        response = await self._rest_api_call(url=url, method=HTTPMethod.POST)
        return UpdateDetails.from_dict(response)


# endregion



"""Models for scheduling."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.json import DataClassJSONMixin

from pylamarzocco.const import SmartStandByType, WeekDay

from ._general import CommandResponse, Thing


@dataclass(kw_only=True)
class WakeUpScheduleSettings(DataClassJSONMixin):
    """Wake up schedule settings."""

    identifier: str | None = field(metadata=field_options(alias="id"), default=None)
    enabled: bool = field(default=False)
    on_time_minutes: int = field(metadata=field_options(alias="onTimeMinutes"))
    off_time_minutes: int = field(metadata=field_options(alias="offTimeMinutes"))
    steam_boiler: bool = field(metadata=field_options(alias="steamBoiler"))
    days: list[WeekDay] = field(default_factory=list)

    class Config(BaseConfig):
        """Config for Mashumaro serialization."""

        serialize_by_alias = True
        omit_none = True


@dataclass(kw_only=True)
class SmartWakeUpSleepSettings(DataClassJSONMixin):
    """Smart wake up sleep settings."""

    smart_stand_by_enabled: bool = field(
        metadata=field_options(alias="smartStandByEnabled"),
        default=False,
    )
    smart_stand_by_minutes: int = field(
        metadata=field_options(alias="smartStandByMinutes"),
        default=0,
    )
    smart_stand_by_minutes_min: int = field(
        metadata=field_options(alias="smartStandByMinutesMin"),
        default=0,
    )
    smart_stand_by_minutes_max: int = field(
        metadata=field_options(alias="smartStandByMinutesMax"),
        default=0,
    )
    smart_stand_by_minutes_step: int = field(
        metadata=field_options(alias="smartStandByMinutesStep"),
        default=0,
    )
    smart_stand_by_after: SmartStandByType = field(
        metadata=field_options(alias="smartStandByAfter"),
        default=SmartStandByType.POWER_ON,
    )
    schedules: list[WakeUpScheduleSettings] = field(
        default_factory=list,
    )
    schedules_dict: dict[str, WakeUpScheduleSettings] = field(
        default_factory=dict,
    )

    @classmethod
    def __post_deserialize__(cls, obj: SmartWakeUpSleepSettings) -> SmartWakeUpSleepSettings:
        # move the firmware to a dict with type as key
        obj.schedules_dict = {schedule.identifier: schedule for schedule in obj.schedules if schedule.identifier}
        return obj



@dataclass(kw_only=True)
class ThingSchedulingSettings(Thing):
    """Scheduling settings."""

    smart_wake_up_sleep_supported: bool = field(
        metadata=field_options(alias="smartWakeUpSleepSupported"),
        default=True,
    )
    smart_wake_up_sleep: SmartWakeUpSleepSettings = field(
        metadata=field_options(alias="smartWakeUpSleep"),
        default_factory=SmartWakeUpSleepSettings,
    )


# ws SUBSCRIBE /ws/sn/SERIAL/scheduling
@dataclass(kw_only=True)
class SmartWakeUpScheduleWebsocketConfig(DataClassJSONMixin):
    """Smart wake up schedule settings from websocket."""

    connected: bool
    auto_stand_by: str | None = field(
        metadata=field_options(alias="autoStandBy"),
        default=None,
    )
    auto_on_off: str | None = field(
        metadata=field_options(alias="autoOnOff"),
        default=None,
    )
    connection_date: datetime | None = field(
        metadata=field_options(
            alias="connectionDate",
            deserialize=lambda ts: datetime.fromtimestamp(ts / 1000, tz=timezone.utc)
            if ts
            else None,
        ),
        default=None,
    )
    auto_stand_by_supported: bool = field(
        metadata=field_options(alias="autoStandBySupported")
    )
    auto_on_off_supported: bool = field(
        metadata=field_options(alias="autoOnOffSupported")
    )
    smart_wake_up_sleep: SmartWakeUpSleepSettings | None = field(
        metadata=field_options(alias="smartWakeUpSleep"),
        default=None,
    )
    weekly_supported: bool = field(metadata=field_options(alias="weeklySupported"))
    smart_wake_up_sleep_supported: bool = field(
        metadata=field_options(alias="smartWakeUpSleepSupported")
    )
    commands: list[CommandResponse] = field(
        default_factory=list,
    )



"""Tests for websocket edge cases and error scenarios."""

import pytest
from unittest.mock import MagicMock, patch, AsyncMock

from pylamarzocco.const import StompMessageType
from pylamarzocco.util import decode_stomp_ws_message, encode_stomp_ws_message
from pylamarzocco.models import WebSocketDetails


class TestWebSocketEdgeCases:
    """Test websocket edge cases and error scenarios."""

    def test_decode_stomp_message_empty_string(self) -> None:
        """Test decoding empty STOMP message."""
        with pytest.raises(ValueError):
            decode_stomp_ws_message("")

    def test_decode_stomp_message_no_double_newline(self) -> None:
        """Test decoding STOMP message without double newline separator."""
        with pytest.raises(ValueError):
            decode_stomp_ws_message("MESSAGE\ndestination:/topic/test")

    def test_decode_stomp_message_invalid_message_type(self) -> None:
        """Test decoding STOMP message with invalid message type."""
        with pytest.raises(ValueError):
            decode_stomp_ws_message("INVALID_TYPE\n\n")

    def test_decode_stomp_message_header_without_colon(self) -> None:
        """Test decoding STOMP message with header line without colon."""
        with pytest.raises(ValueError):
            decode_stomp_ws_message("MESSAGE\ninvalid_header_line\n\nbody")

    def test_encode_stomp_message_special_characters(self) -> None:
        """Test encoding STOMP message with special characters."""
        headers = {"destination": "/topic/test", "content-type": "application/json"}
        body = '{"message": "Hello\\nWorld\\tWith\\rSpecial\\x00Chars"}'
        
        result = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, body)
        
        # Should contain all parts
        assert "MESSAGE" in result
        assert "destination:/topic/test" in result
        assert "content-type:application/json" in result
        assert body in result
        assert result.endswith("\x00")

    def test_encode_stomp_message_unicode_content(self) -> None:
        """Test encoding STOMP message with unicode content."""
        headers = {"destination": "/topic/test"}
        body = '{"message": "Hello 世界 🌍"}'
        
        result = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, body)
        
        # Should handle unicode properly
        assert body in result
        
        # Test round trip
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(result)
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == body

    def test_websocket_details_edge_cases(self) -> None:
        """Test WebSocketDetails edge cases."""
        # Test with None websocket
        details = WebSocketDetails(ws=None)
        assert not details.connected
        
        # Test with closed websocket
        mock_ws = MagicMock()
        mock_ws.closed = True
        details = WebSocketDetails(ws=mock_ws)
        assert not details.connected
        
        # Test state change
        mock_ws.closed = False
        assert details.connected

    async def test_websocket_details_disconnect_with_none_callback(self) -> None:
        """Test WebSocketDetails disconnect with None callback."""
        details = WebSocketDetails(ws=None, disconnect_callback=None)
        
        # Should not raise
        await details.disconnect()

    async def test_websocket_details_disconnect_callback_exception(self) -> None:
        """Test WebSocketDetails disconnect when callback raises exception."""
        def failing_callback():
            raise Exception("Callback failed")
        
        details = WebSocketDetails(disconnect_callback=failing_callback)
        
        # Should not suppress the exception
        with pytest.raises(Exception, match="Callback failed"):
            await details.disconnect()

    def test_stomp_message_types_coverage(self) -> None:
        """Test all STOMP message types can be encoded/decoded."""
        test_headers = {"test": "header"}
        test_body = "test body"
        
        # Test each message type
        for msg_type in StompMessageType:
            encoded = encode_stomp_ws_message(msg_type, test_headers, test_body)
            decoded_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded)
            
            assert decoded_type == msg_type
            assert decoded_headers == test_headers
            assert decoded_body == test_body

    def test_encode_stomp_message_large_headers(self) -> None:
        """Test encoding STOMP message with many headers."""
        headers = {f"header_{i}": f"value_{i}" for i in range(100)}
        body = "test body"
        
        result = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, body)
        
        # Should handle large number of headers
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(result)
        
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == body

    def test_encode_stomp_message_long_header_values(self) -> None:
        """Test encoding STOMP message with very long header values."""
        long_value = "x" * 10000
        headers = {"destination": "/topic/test", "long_header": long_value}
        body = "test body"
        
        result = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, body)
        
        # Should handle long header values
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(result)
        
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == body

    def test_decode_stomp_message_large_body(self) -> None:
        """Test decoding STOMP message with large body."""
        headers = {"destination": "/topic/test"}
        large_body = "x" * 100000  # 100KB body
        
        encoded = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, large_body)
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded)
        
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == large_body

    def test_encode_decode_empty_body(self) -> None:
        """Test encoding/decoding STOMP message with empty body."""
        headers = {"destination": "/topic/test"}
        
        # Test with None body
        encoded_none = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, None)
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded_none)
        
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == ""
        
        # Test with empty string body
        encoded_empty = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, "")
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded_empty)
        
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == ""

    def test_header_with_multiple_colons(self) -> None:
        """Test header values containing multiple colons."""
        headers = {
            "destination": "/topic/test",
            "authorization": "Bearer token:with:multiple:colons:here",
            "url": "https://example.com:8080/path"
        }
        body = "test"
        
        encoded = encode_stomp_ws_message(StompMessageType.MESSAGE, headers, body)
        msg_type, decoded_headers, decoded_body = decode_stomp_ws_message(encoded)
        
        assert msg_type == StompMessageType.MESSAGE
        assert decoded_headers == headers
        assert decoded_body == body

    def test_websocket_connection_state_transitions(self) -> None:
        """Test WebSocketDetails connection state transitions."""
        mock_ws = MagicMock()
        details = WebSocketDetails(ws=mock_ws)
        
        # Initially open
        mock_ws.closed = False
        assert details.connected
        
        # Close it
        mock_ws.closed = True
        assert not details.connected
        
        # Reopen (simulating reconnection)
        mock_ws.closed = False
        assert details.connected


"""General utility functions."""

from aiohttp import ClientResponse

def is_success(response: ClientResponse) -> bool:
    """Check if response is successful."""
    return 200 <= response.status < 300


"""Testing models serialization and deserialization."""

import logging

import pytest
from syrupy import SnapshotAssertion

from pylamarzocco.const import WidgetType
from pylamarzocco.models import (
    SmartWakeUpScheduleWebsocketConfig,
    ThingDashboardWebsocketConfig,
)
from pylamarzocco.models._statistics import ThingStatistics

from .conftest import load_fixture


async def test_device_config(snapshot: SnapshotAssertion) -> None:
    """Test the config model serialization."""

    fixture = load_fixture("machine", "config_micra.json")
    device = ThingDashboardWebsocketConfig.from_dict(fixture)
    assert device.to_dict() == snapshot


async def test_scheduling_websocket_config(snapshot: SnapshotAssertion) -> None:
    """Test the config model serialization."""

    fixture = load_fixture("machine", "schedule_ws.json")
    schedule = SmartWakeUpScheduleWebsocketConfig.from_dict(fixture)
    assert schedule.to_dict() == snapshot


async def test_removed_widgets_with_unknown_codes(caplog: pytest.LogCaptureFixture) -> None:
    """Test that removed widgets with unknown codes are filtered out and logged."""
    config_data = {
        "connected": True,
        "removedWidgets": [
            {"code": "NewWidget2026", "index": 1},         # Unknown - will be filtered
            {"code": "UnknownFutureWidget", "index": 2},   # Unknown - will be filtered
            {"code": "CMMachineStatus", "index": 3},       # Known - will be kept
        ],
        "connectionDate": 1742439620235,
        "widgets": [],
        "uuid": "test-uuid",
        "commands": [],
    }
    
    with caplog.at_level(logging.WARNING):
        config = ThingDashboardWebsocketConfig.from_dict(config_data)
    
    # Only the known widget should remain
    assert len(config.removed_widgets) == 1
    assert config.removed_widgets[0].code == WidgetType.CM_MACHINE_STATUS
    assert isinstance(config.removed_widgets[0].code, WidgetType)
    
    # Verify warnings were logged for unknown codes
    assert "NewWidget2026" in caplog.text
    assert "UnknownFutureWidget" in caplog.text


async def test_widgets_with_unknown_codes(caplog: pytest.LogCaptureFixture) -> None:
    """Test that active widgets with unknown codes are filtered out and logged."""
    config_data = {
        "connected": True,
        "removedWidgets": [],
        "connectionDate": 1742439620235,
        "widgets": [
            {
                "code": "UnknownWidget",
                "index": 1,
                "output": {"widget_type": "UnknownWidget"},
            },
            {
                "code": "CMMachineStatus",
                "index": 2,
                "output": {
                    "widget_type": "CMMachineStatus",
                    "status": "StandBy",
                    "availableModes": ["BrewingMode", "StandBy"],
                    "mode": "StandBy",
                    "nextStatus": None,
                    "brewingStartTime": None,
                },
            },
        ],
        "uuid": "test-uuid",
        "commands": [],
    }
    
    with caplog.at_level(logging.WARNING):
        config = ThingDashboardWebsocketConfig.from_dict(config_data)
    
    # Only the known widget should remain
    assert len(config.widgets) == 1
    assert config.widgets[0].code == WidgetType.CM_MACHINE_STATUS
    
    # Verify warning was logged for unknown code and includes JSON
    assert "UnknownWidget" in caplog.text
    # Verify that full JSON is logged for dict objects
    assert '"index": 1' in caplog.text


async def test_widget_code_strings_filtering(caplog: pytest.LogCaptureFixture) -> None:
    """Test that widget code strings (not dicts) are filtered properly."""
    stats_data = {
        "serialNumber": "test-serial",
        "selectedWidgetCodes": [
            "CMMachineStatus",     # Known - will be kept
            "UnknownCode1",        # Unknown - will be filtered
            "CMCoffeeBoiler",      # Known - will be kept
            "UnknownCode2",        # Unknown - will be filtered
        ],
        "allWidgetCodes": [
            "CMSteamBoilerLevel",  # Known - will be kept
            "FutureWidget",        # Unknown - will be filtered
        ],
        "selectedWidgets": [],
    }
    
    with caplog.at_level(logging.WARNING):
        stats = ThingStatistics.from_dict(stats_data)
    
    # Only known codes should remain
    assert len(stats.selected_widget_codes) == 2
    assert WidgetType.CM_MACHINE_STATUS in stats.selected_widget_codes
    assert WidgetType.CM_COFFEE_BOILER in stats.selected_widget_codes
    
    assert len(stats.all_widget_codes) == 1
    assert WidgetType.CM_STEAM_BOILER_LEVEL in stats.all_widget_codes
    
    # Verify warnings were logged for unknown codes (strings, not dicts)
    assert "UnknownCode1" in caplog.text
    assert "UnknownCode2" in caplog.text
    assert "FutureWidget" in caplog.text


async def test_mixed_valid_invalid_widgets(caplog: pytest.LogCaptureFixture) -> None:
    """Test handling of mix of valid and invalid widgets."""
    config_data = {
        "connected": True,
        "removedWidgets": [
            {"code": "CMMachineStatus", "index": 1},
            {"code": "InvalidWidget1", "index": 2},
            {"code": "CMCoffeeBoiler", "index": 3},
            {"code": "InvalidWidget2", "index": 4},
            {"code": "CMSteamBoilerLevel", "index": 5},
        ],
        "connectionDate": 1742439620235,
        "widgets": [],
        "uuid": "test-uuid",
        "commands": [],
    }
    
    with caplog.at_level(logging.WARNING):
        config = ThingDashboardWebsocketConfig.from_dict(config_data)
    
    # Should have 3 valid widgets
    assert len(config.removed_widgets) == 3
    assert config.removed_widgets[0].code == WidgetType.CM_MACHINE_STATUS
    assert config.removed_widgets[1].code == WidgetType.CM_COFFEE_BOILER
    assert config.removed_widgets[2].code == WidgetType.CM_STEAM_BOILER_LEVEL
    
    # Verify both invalid widgets were logged with full JSON
    assert "InvalidWidget1" in caplog.text
    assert "InvalidWidget2" in caplog.text
    assert '"index": 2' in caplog.text
    assert '"index": 4' in caplog.text


async def test_empty_widget_lists() -> None:
    """Test that empty widget lists are handled correctly."""
    config_data = {
        "connected": True,
        "removedWidgets": [],
        "connectionDate": 1742439620235,
        "widgets": [],
        "uuid": "test-uuid",
        "commands": [],
    }
    
    config = ThingDashboardWebsocketConfig.from_dict(config_data)
    
    assert len(config.removed_widgets) == 0
    assert len(config.widgets) == 0


async def test_all_widgets_invalid(caplog: pytest.LogCaptureFixture) -> None:
    """Test that when all widgets are invalid, result is empty list."""
    config_data = {
        "connected": True,
        "removedWidgets": [
            {"code": "InvalidWidget1", "index": 1},
            {"code": "InvalidWidget2", "index": 2},
            {"code": "InvalidWidget3", "index": 3},
        ],
        "connectionDate": 1742439620235,
        "widgets": [],
        "uuid": "test-uuid",
        "commands": [],
    }
    
    with caplog.at_level(logging.WARNING):
        config = ThingDashboardWebsocketConfig.from_dict(config_data)
    
    # All widgets should be filtered out
    assert len(config.removed_widgets) == 0
    
    # All three should have been logged
    assert "InvalidWidget1" in caplog.text
    assert "InvalidWidget2" in caplog.text
    assert "InvalidWidget3" in caplog.text
